/*
 * aepsilon.cpp
 *
 *  Created on: 24. 2. 2014
 *      Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>
#include <sax/FromXMLParserHelper.h>

#include <factory/XmlDataFactory.hpp>
#include <exception/CommonException.h>

#include "automaton/simplify/EpsilonRemoverIncoming.h"
#include "automaton/simplify/EpsilonRemoverOutgoing.h"
#include "automaton/simplify/efficient/EpsilonRemoverIncoming.h"
#include "automaton/simplify/efficient/EpsilonRemoverOutgoing.h"
#include "grammar/simplify/EpsilonRemover.h"

int main(int argc, char** argv) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd("Automaton rename binary", ' ', "0.01");

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Automaton/Grammar where to remove epsilon transitions",		false,	"-",		"file");
		cmd.add( input );

		std::vector<std::string> algorithms{"incoming", "outgoing", "default" };
		TCLAP::ValuesConstraint<std::string> allowedAlgorithms( algorithms );
		TCLAP::ValueArg<std::string> algorithm(	"a",	"algorithm",		"Specifies algorithm to use",		false,	"default",	&allowedAlgorithms);
		cmd.add( algorithm );

		TCLAP::SwitchArg efficient(		"e",	"efficient",		"Use efficient implementation",		false);
		cmd.add( efficient );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		measurements::start("Overal", measurements::Type::OVERALL);
		measurements::start("Input read", measurements::Type::AUXILIARY);

		std::deque<sax::Token> tokens = sax::FromXMLParserHelper::parseInput(input);

		if(algorithm.getValue() == "outgoing") {
			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(std::move(tokens));

			measurements::end();
			measurements::start("Algorithm", measurements::Type::MAIN);

			if(efficient.getValue()) {
				automaton::Automaton res = automaton::simplify::efficient::EpsilonRemoverOutgoing::remove(automaton);

				measurements::end();
				measurements::start("Output write", measurements::Type::AUXILIARY);

				alib::XmlDataFactory::toStdout(res);
			} else {
				automaton::Automaton res = automaton::simplify::EpsilonRemoverOutgoing::remove(automaton);

				measurements::end();
				measurements::start("Output write", measurements::Type::AUXILIARY);

				alib::XmlDataFactory::toStdout(res);
			}
		} else if(algorithm.getValue() == "incoming") {
			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(std::move(tokens));

			measurements::end();
			measurements::start("Algorithm", measurements::Type::MAIN);

			if(efficient.getValue()) {
				automaton::Automaton res = automaton::simplify::efficient::EpsilonRemoverIncoming::remove(automaton);

				measurements::end();
				measurements::start("Output write", measurements::Type::AUXILIARY);

				alib::XmlDataFactory::toStdout(res);
			} else {
				automaton::Automaton res = automaton::simplify::EpsilonRemoverIncoming::remove(automaton);

				measurements::end();
				measurements::start("Output write", measurements::Type::AUXILIARY);

				alib::XmlDataFactory::toStdout(res);
			}
		} else if(algorithm.getValue() == "default") {
			if(alib::XmlDataFactory::first<automaton::Automaton>(tokens)) {
				automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(std::move(tokens));

				measurements::end();
				measurements::start("Algorithm", measurements::Type::MAIN);

				if(efficient.getValue()) {
					automaton::Automaton res = automaton::simplify::EpsilonRemoverIncoming::remove(automaton);

					measurements::end();
					measurements::start("Output write", measurements::Type::AUXILIARY);

					alib::XmlDataFactory::toStdout(res);
				} else {
					automaton::Automaton res = automaton::simplify::efficient::EpsilonRemoverIncoming::remove(automaton);

					measurements::end();
					measurements::start("Output write", measurements::Type::AUXILIARY);

					alib::XmlDataFactory::toStdout(res);
				}
			} else if(alib::XmlDataFactory::first<grammar::Grammar>(tokens)) {
				grammar::Grammar grammar = alib::XmlDataFactory::fromTokens<grammar::Grammar>(std::move(tokens));

				measurements::end();
				measurements::start("Algorithm", measurements::Type::MAIN);

				grammar::Grammar res = grammar::simplify::EpsilonRemover::remove(grammar);

				measurements::end();
				measurements::start("Output write", measurements::Type::AUXILIARY);

				alib::XmlDataFactory::toStdout(res);
			} else
				throw exception::CommonException("Invalid formalism");
		} else {
			throw exception::CommonException("Invalid algorithm");
		}

		measurements::end();
		measurements::end();

		if(measure.getValue()) std::cmeasure << measurements::results() << std::endl;

		return 0;
	} catch (const exception::CommonException& exception) {
		alib::XmlDataFactory::toStdout(exception);
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
