/*
 * Minimize.cpp
 *
 *  Created on: Dec 9, 2013
 *	  Author: honza
 */

#include "Minimize.h"

#include <map>
#include <set>
#include <iomanip>
#include <sstream>

#include <alphabet/Symbol.h>

#include <global/GlobalData.h>

namespace automaton {

namespace simplify {

automaton::Automaton Minimize::minimize(const automaton::Automaton& automaton) {
	return dispatch(automaton.getData());
}

automaton::DFA Minimize::minimize(const automaton::DFA& dfa) {
	if(dfa.getFinalStates().size() == 0) {
		automaton::DFA result(automaton::State(0));
		result.setInputAlphabet(dfa.getInputAlphabet());
		return result;
	}

	std::map<automaton::State, std::map<alphabet::Symbol, automaton::State > > refactor;

	for(const automaton::State& state : dfa.getStates()) {
		refactor.insert(std::make_pair(state, std::map<alphabet::Symbol, automaton::State>()));
	}

	for(const std::pair<const std::pair<automaton::State, alphabet::Symbol>, automaton::State>& transition : dfa.getTransitions()) {
		refactor[transition.first.first].insert(std::make_pair(transition.first.second, transition.second));
	}

	std::map<automaton::State, automaton::State> toEquvivalentStates1; //original state to equivalent state
	std::map<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> > minimizedTransitionFunction1; //mapped to the original state

	std::map<automaton::State, automaton::State> toEquvivalentStates2;
	std::map<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> > minimizedTransitionFunction2;

	const automaton::State *firstFinal = NULL, *firstNonfinal = NULL;
	for(const automaton::State& state : dfa.getStates()) {
		if(dfa.getFinalStates().count(state) == 0) { // not a final state
			if(!firstNonfinal) firstNonfinal = &state;
			toEquvivalentStates2.insert(std::pair<automaton::State, automaton::State>(state, *firstNonfinal));
		} else {
			if(!firstFinal) firstFinal = &state;
			toEquvivalentStates2.insert(std::pair<automaton::State, automaton::State>(state, *firstFinal));
		}
	}

	for(const std::pair<const automaton::State, std::map<alphabet::Symbol, automaton::State> >& transition : refactor) {
		const automaton::State& from = toEquvivalentStates2.find(transition.first)->second;
		std::set<std::pair<alphabet::Symbol, automaton::State> > transitionFunction;

		for(const std::pair<const alphabet::Symbol, automaton::State> & transitionFromState : transition.second) {
			transitionFunction.insert(std::make_pair(transitionFromState.first, toEquvivalentStates2.find(transitionFromState.second)->second));
		}

		minimizedTransitionFunction2[std::make_pair(from, transitionFunction)].insert(transition.first);
	}

	size_t delta_iter = 0;
	if(common::GlobalData::verbose) {
		print_progress(dfa, minimizedTransitionFunction2, delta_iter++);
	}

	do {
		toEquvivalentStates1 = toEquvivalentStates2;
		minimizedTransitionFunction1 = minimizedTransitionFunction2;

		toEquvivalentStates2.clear();
		minimizedTransitionFunction2.clear();

		for(const std::pair<const std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> >& transition : minimizedTransitionFunction1) {
			for(const automaton::State& target : transition.second) {
				toEquvivalentStates2.insert(std::make_pair(target, *(transition.second.begin())));
			}
		}

		for(const std::pair<const automaton::State, std::map<alphabet::Symbol, automaton::State> >& transition : refactor) {
			const automaton::State& from = toEquvivalentStates2.find(transition.first)->second;
			std::set<std::pair<alphabet::Symbol, automaton::State> > transitionFunction;

			for(const std::pair<const alphabet::Symbol, automaton::State> & transitionFromState : transition.second) {
				transitionFunction.insert(std::make_pair(transitionFromState.first, toEquvivalentStates2.find(transitionFromState.second)->second));
			}

			minimizedTransitionFunction2[std::make_pair(from, transitionFunction)].insert(transition.first);
		}

		if(common::GlobalData::verbose) {
			print_progress(dfa, minimizedTransitionFunction2, delta_iter++);
		}

	} while(minimizedTransitionFunction1.size() != minimizedTransitionFunction2.size());

	const automaton::State* initialState = NULL;
	for(const std::pair<const std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> >& transition : minimizedTransitionFunction2) {
		if(transition.second.count(dfa.getInitialState())) {
			initialState = &(transition.first.first);
			break;
		}
	}

	automaton::DFA result(*initialState);

	result.setInputAlphabet(dfa.getInputAlphabet());

	for(const std::pair<const std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> >& transition : minimizedTransitionFunction2) {
		result.addState(transition.first.first);
		if(dfa.getFinalStates().count(*(transition.second.begin()))) {
			result.addFinalState(transition.first.first);
		}
	}

	for(const std::pair<const std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> >& transition : minimizedTransitionFunction2) {
		for(const std::pair<alphabet::Symbol, automaton::State>& transitionFromState : transition.first.second) {
			result.addTransition(transition.first.first, transitionFromState.first, transitionFromState.second);
		}
	}

	return result;
}

#define RESETSS(x) {(x).clear(); (x).str("");}

void Minimize::print_progress(const automaton::DFA& dfa, const std::map<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> >& minimizedTransitionFunction, size_t iter) {
	std::clog << "delta " << iter << std::endl;

	//std::map<std::pair<automaton::State, std::set<std::pair<alphabet::Symbol, automaton::State> > >, std::set<automaton::State> > minimizedTransitionFunction1; //mapped to the original state

	/* need to restruct this first so we have table like:	orig state | new state | trans_symb_1 | trans_symb_2 | ... | trans_symb_n */
	// we surely have DFA here (transition map hence)
	std::map<std::pair<automaton::State, automaton::State>, std::map<alphabet::Symbol, automaton::State>> printMap;
	for(const auto& kv: minimizedTransitionFunction) {
		for(const auto& state : kv.second) {
			std::map<alphabet::Symbol, automaton::State> stateTransMap;
			for(const auto& transition : kv.first.second) {
				stateTransMap.insert(std::make_pair(transition.first, transition.second));
			}
			printMap.insert(std::make_pair(std::make_pair(state, kv.first.first), stateTransMap));
		}
	}

	size_t stateWidth = 1, stateMapWidth = 1;
	std::map<alphabet::Symbol, size_t> colWidths;
	std::ostringstream ss;

	for(const alphabet::Symbol& symbol : dfa.getInputAlphabet()) {
		ss << symbol;
		colWidths[symbol] = ss.str().size();
		RESETSS(ss);
	}
	for(const auto& kv : printMap) {
		ss << kv.first.first;
		stateWidth = std::max(stateWidth, ss.str().size());
		RESETSS(ss);

		ss << kv.first.second;
		stateMapWidth = std::max(stateMapWidth, ss.str().size());
		RESETSS(ss);

		for(const alphabet::Symbol& symbol : dfa.getInputAlphabet()) {
			auto it = kv.second.find(symbol);
			if(it != kv.second.end())
			{
				ss << it -> second;
				colWidths[symbol] = std::max(colWidths[symbol], ss.str().size());
				RESETSS(ss);
			}
		}
	}

	std::clog << std::setw(stateWidth) << "";
	std::clog << " | ";
	std::clog << std::setw(stateMapWidth) << "";
	std::clog << " | ";
	for(const alphabet::Symbol& symbol : dfa.getInputAlphabet()) {
		ss << symbol;
		std::clog << std::setw(colWidths[symbol]) << ss.str() << " | ";
		RESETSS(ss);
	}
	std::clog << std::endl;
	for(const auto& kv : printMap) {
		ss << kv.first.first;
		std::clog << std::setw(stateWidth) << ss.str() << " | ";
		RESETSS(ss);

		ss << kv.first.second;
		std::clog << std::setw(stateMapWidth) << ss.str() << " | ";
		RESETSS(ss);

		for(const auto& symbol : dfa.getInputAlphabet()) {
			auto it = kv.second.find(symbol);
			if(it != kv.second.end())
			{
				ss << it -> second;
				std::clog << std::setw(colWidths[symbol]) << ss.str();
				RESETSS(ss);
			}
			else
			{
				std::clog << std::setw(colWidths[symbol]) << "";
			}
			std::clog << " | ";
		}
		std::clog << std::endl;
	}
	std::clog << std::endl;
}

auto MinimizeNFA = Minimize::RegistratorWrapper<automaton::DFA, automaton::DFA>( Minimize::minimize );

} /* namespace simplify */

} /* namespace automaton */
