/*
 * AutomataConcatenation.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataConcatenation.h"
#include <label/Label.h>
#include "common/PairLabel.h"

#define AUTOMATON_FIRST  (label::labelFrom(1))
#define AUTOMATON_SECOND (label::labelFrom(2))

namespace automaton {

namespace transform {

automaton::Automaton AutomataConcatenation::concatenation(const automaton::Automaton& first, const automaton::Automaton& second) {
	return dispatch(first.getData(), second.getData());
}

automaton::NFA AutomataConcatenation::concatenation(const automaton::NFA& first, const automaton::NFA& second) {
	automaton::State q01q02(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::NFA res(pairLabel(AUTOMATON_FIRST, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(pairLabel(AUTOMATON_FIRST, q));
	for(const auto& q : second.getStates())
		res.addState(pairLabel(AUTOMATON_SECOND, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions()) {
		for(const auto& q : t.second) {
			res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_FIRST, q));

			if(first.getFinalStates().count(q) > 0)
				res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, second.getInitialState()));
		}
	}
	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(AUTOMATON_SECOND, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, q));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_FIRST, q));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_SECOND, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(AUTOMATON_SECOND, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

auto AutomataConcatenationNFA = AutomataConcatenation::RegistratorWrapper<automaton::NFA, automaton::NFA>(AutomataConcatenation::concatenation);

automaton::NFA AutomataConcatenation::concatenation(const automaton::DFA& first, const automaton::DFA& second) {
	automaton::State q01q02(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::NFA res(pairLabel(AUTOMATON_FIRST, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(pairLabel(AUTOMATON_FIRST, q));
	for(const auto& q : second.getStates())
		res.addState(pairLabel(AUTOMATON_SECOND, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions()) {
		res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_FIRST, t.second));

		if(first.getFinalStates().count(t.second) > 0)
			res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, second.getInitialState()));
	}
	for(const auto& t : second.getTransitions())
		res.addTransition(pairLabel(AUTOMATON_SECOND, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, t.second));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_FIRST, t.second));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_SECOND, t.second));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(AUTOMATON_SECOND, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

auto AutomataConcatenationDFA = AutomataConcatenation::RegistratorWrapper<automaton::NFA, automaton::DFA>(AutomataConcatenation::concatenation);

} /* namespace transform */

} /* namespace automaton */
