/*
 * SLR1ParseTable.cpp
 *
 *  Created on: 4. 5. 2016
 *	  Author: Martin Kocicka
 */

#include "SLR1ParseTable.h"

#include "LRParser.h"
#include "LR0Parser.h"
#include "Follow.h"

#include <exception/CommonException.h>
#include <label/LR0ItemsLabel.h>

namespace grammar {

namespace parsing {

void SLR1ParseTable::insertToActionTable ( LRActionTable & actionTable, LRActionTable::key_type key, LRActionTable::mapped_type value ) {
	std::pair<LRActionTable::iterator, bool> result = actionTable.insert ( { key, value } );
	if ( ! result.second ) {
		throw exception::CommonException ( "Grammar is not SLR(1)!" );
	}
}

LRActionTable SLR1ParseTable::getActionTable ( grammar::CFG originalGrammar ) {
	LRActionTable actionTable;
	grammar::CFG augmentedGrammar = LRParser::getAugmentedGrammar ( originalGrammar );
	automaton::DFA parsingAutomaton = LR0Parser::getAutomaton ( originalGrammar );
	for ( const automaton::State & state : parsingAutomaton.getStates ( ) ) {
		LR0Items items = static_cast < const label::LR0ItemsLabel & > ( state.getName ( ) . getData ( ) ) . getItems ( );
		std::map < std::pair < automaton::State, alphabet::Symbol >, automaton::State > transitionsFromCurrentState = parsingAutomaton.getTransitionsFromState ( state );
		for ( const LR0Items::value_type & nonterminalItems : items ) {
			alphabet::Symbol leftHandSide = nonterminalItems.first;
			for ( const std::pair < unsigned, std::vector < alphabet::Symbol > > & item : nonterminalItems.second ) {
				unsigned position = item.first;
				std::vector < alphabet::Symbol > rightHandSide = item.second;

				if ( position == rightHandSide.size ( ) ) {
					if ( leftHandSide == augmentedGrammar.getInitialSymbol ( ) ) {
						insertToActionTable(actionTable, { state, LRParser::getEndOfInputSymbol(augmentedGrammar) }, { LRAction::Accept, state } );
						continue;
					}

					grammar::parsing::FollowResult2 followSet = grammar::parsing::Follow::follow ( augmentedGrammar, leftHandSide );
					std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > currentRule = { leftHandSide, rightHandSide };
					for ( const std::variant < alphabet::Symbol, string::Epsilon > & followSymbol : followSet ) {
						if ( followSymbol.is < alphabet::Symbol > ( ) ) {
							insertToActionTable(actionTable, { state, followSymbol.get < alphabet::Symbol > ( ) }, { LRAction::Reduce, currentRule } );
						} else {
							insertToActionTable(actionTable, { state, LRParser::getEndOfInputSymbol(augmentedGrammar) }, { LRAction::Reduce, currentRule } );
						}
					}
					continue;
				}

				alphabet::Symbol currentSymbol = rightHandSide[position];
				if ( originalGrammar.getTerminalAlphabet ( ) . find ( currentSymbol ) != originalGrammar.getTerminalAlphabet ( ) . end ( ) ) {
					insertToActionTable(actionTable, { state, currentSymbol }, { LRAction::Shift, transitionsFromCurrentState.find ( { state, currentSymbol } )->second } );
				}
			}
		}
	}

	return actionTable;
}

LRGotoTable SLR1ParseTable::getGotoTable ( grammar::CFG originalGrammar ) {
	LRGotoTable gotoTable;
	grammar::CFG augmentedGrammar = LRParser::getAugmentedGrammar ( originalGrammar );
	automaton::DFA parsingAutomaton = LR0Parser::getAutomaton ( originalGrammar );
	for ( const automaton::State & state : parsingAutomaton.getStates ( ) ) {
		std::map < std::pair < automaton::State, alphabet::Symbol >, automaton::State > transitionsFromCurrentState = parsingAutomaton.getTransitionsFromState ( state );
		for ( const alphabet::Symbol & nonterminal : augmentedGrammar.getNonterminalAlphabet ( ) ) {
			std::map < std::pair<automaton::State, alphabet::Symbol >, automaton::State >::iterator transitionIterator = transitionsFromCurrentState.find ( { state, nonterminal } );
			if ( transitionIterator != transitionsFromCurrentState.end ( ) ) {
				gotoTable.insert ( { { state, nonterminal }, transitionIterator->second } );
			}
		}
	}

	return gotoTable;
}

} /* namespace parsing */

} /* namespace grammar */
