/*
 * GlushkovTraversal.cpp
 *
 *  Created on: 13. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "GlushkovTraversal.h"

#include "../properties/RegExpEpsilon.h"
#include <alphabet/SymbolPairSymbol.h>

#include <exception/CommonException.h>
#include <iterator>

namespace regexp {


alphabet::Symbol GlushkovTraversal::getSymbolFromGlushkovPair(const alphabet::Symbol & symbol) {
	const alphabet::SymbolPairSymbol & symbolPair = (const alphabet::SymbolPairSymbol&) symbol.getData();
	return symbolPair.getData().first;
}

bool GlushkovTraversal::pos ( const UnboundedRegExpSymbol & symbol, const regexp::UnboundedRegExp & node ) {
	return pos ( node.getRegExp ( ), symbol );
}

std::set < UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExp & re ) {
	return first ( re.getRegExp ( ) );
}

std::set < UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExp & re ) {
	return last ( re.getRegExp ( ) );
}

std::set < UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExp & re, const UnboundedRegExpSymbol & symbol ) {
	return follow ( re.getRegExp ( ), symbol );
}

// -----------------------------------------------------------------------------

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExpElement & node ) {
	const regexp::UnboundedRegExpAlternation * alternation = dynamic_cast < const regexp::UnboundedRegExpAlternation * > ( & node );
	const regexp::UnboundedRegExpConcatenation * concatenation = dynamic_cast < const regexp::UnboundedRegExpConcatenation * > ( & node );
	const regexp::UnboundedRegExpIteration * iteration = dynamic_cast < const regexp::UnboundedRegExpIteration * > ( & node );
	const regexp::UnboundedRegExpSymbol * symbol = dynamic_cast < const regexp::UnboundedRegExpSymbol * > ( & node );
	const regexp::UnboundedRegExpEmpty * empty = dynamic_cast < const regexp::UnboundedRegExpEmpty * > ( & node );
	const regexp::UnboundedRegExpEpsilon * eps = dynamic_cast < const regexp::UnboundedRegExpEpsilon * > ( & node );

	if ( alternation )
		return first ( * alternation );
	else if ( concatenation )
		return first ( * concatenation );
	else if ( iteration )
		return first ( * iteration );
	else if ( eps )
		return first ( * eps );
	else if ( empty )
		return first ( * empty );
	else if ( symbol )
		return first ( * symbol );

	throw exception::CommonException ( "GlushkovTraversal::first - invalid RegExpElement node" );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExpAlternation & node ) {
	std::set < regexp::UnboundedRegExpSymbol > ret, tmp;

	for ( auto const & element : node.getElements ( ) ) {
		tmp = first ( * element );
		ret.insert ( tmp.begin ( ), tmp.end ( ) );
	}

	return ret;
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExpConcatenation & node ) {
	std::set < regexp::UnboundedRegExpSymbol > ret, tmp;

	for ( auto const & element : node.getElements ( ) ) {
		tmp = first ( * element );
		ret.insert ( tmp.begin ( ), tmp.end ( ) );

		if ( !regexp::properties::RegExpEpsilon::languageContainsEpsilon ( * element ) ) // If regexp of this subtree can match epsilon, then we need to add next subtree
			break;
	}

	return ret;
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExpIteration & node ) {
	return first ( node.getElement ( ) );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExpSymbol & node ) {
	return std::set < regexp::UnboundedRegExpSymbol > { node };
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExpEpsilon & /* node */ ) {
	return std::set < regexp::UnboundedRegExpSymbol > ( );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::first ( const regexp::UnboundedRegExpEmpty & /* node */ ) {
	return std::set < regexp::UnboundedRegExpSymbol > ( );
}

// ----------------------------------------------------------------------------

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExpElement & node ) {
	const regexp::UnboundedRegExpAlternation * alternation = dynamic_cast < const regexp::UnboundedRegExpAlternation * > ( & node );
	const regexp::UnboundedRegExpConcatenation * concatenation = dynamic_cast < const regexp::UnboundedRegExpConcatenation * > ( & node );
	const regexp::UnboundedRegExpIteration * iteration = dynamic_cast < const regexp::UnboundedRegExpIteration * > ( & node );
	const regexp::UnboundedRegExpSymbol * symbol = dynamic_cast < const regexp::UnboundedRegExpSymbol * > ( & node );
	const regexp::UnboundedRegExpEmpty * empty = dynamic_cast < const regexp::UnboundedRegExpEmpty * > ( & node );
	const regexp::UnboundedRegExpEpsilon * eps = dynamic_cast < const regexp::UnboundedRegExpEpsilon * > ( & node );

	if ( symbol )
		return last ( * symbol );
	else if ( alternation )
		return last ( * alternation );
	else if ( concatenation )
		return last ( * concatenation );
	else if ( iteration )
		return last ( * iteration );
	else if ( eps )
		return last ( * eps );
	else if ( empty )
		return last ( * empty );

	throw exception::CommonException ( "GlushkovTraversal::last - invalid RegExpElement node" );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExpAlternation & node ) {
	std::set < regexp::UnboundedRegExpSymbol > ret;

	for ( const auto & element : node.getElements ( ) ) {
		std::set < regexp::UnboundedRegExpSymbol > tmp = last ( * element );
		ret.insert ( tmp.begin ( ), tmp.end ( ) );
	}

	return ret;
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExpConcatenation & node ) {
	std::set < regexp::UnboundedRegExpSymbol > ret;

	for ( const auto & element : std::make_reverse ( node.getElements ( ) ) ) {
		std::set < regexp::UnboundedRegExpSymbol > tmp = last ( * element );
		ret.insert ( tmp.begin ( ), tmp.end ( ) );

		if ( !regexp::properties::RegExpEpsilon::languageContainsEpsilon ( * element ) )
			break;
	}

	return ret;
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExpIteration & node ) {
	return last ( node.getElement ( ) );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExpSymbol & node ) {
	return std::set < regexp::UnboundedRegExpSymbol > { node };
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExpEpsilon & /* node */ ) {
	return std::set < regexp::UnboundedRegExpSymbol > ( );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::last ( const regexp::UnboundedRegExpEmpty & /* node */ ) {
	return std::set < regexp::UnboundedRegExpSymbol > ( );
}

// ----------------------------------------------------------------------------

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExpElement & node, const regexp::UnboundedRegExpSymbol & symbolptr ) {
	const regexp::UnboundedRegExpAlternation * alternation = dynamic_cast < const regexp::UnboundedRegExpAlternation * > ( & node );
	const regexp::UnboundedRegExpConcatenation * concatenation = dynamic_cast < const regexp::UnboundedRegExpConcatenation * > ( & node );
	const regexp::UnboundedRegExpIteration * iteration = dynamic_cast < const regexp::UnboundedRegExpIteration * > ( & node );
	const regexp::UnboundedRegExpSymbol * symbol = dynamic_cast < const regexp::UnboundedRegExpSymbol * > ( & node );
	const regexp::UnboundedRegExpEmpty * empty = dynamic_cast < const regexp::UnboundedRegExpEmpty * > ( & node );
	const regexp::UnboundedRegExpEpsilon * eps = dynamic_cast < const regexp::UnboundedRegExpEpsilon * > ( & node );

	if ( alternation )
		return follow ( * alternation, symbolptr );

	else if ( concatenation )
		return follow ( * concatenation, symbolptr );

	else if ( iteration )
		return follow ( * iteration, symbolptr );

	else if ( symbol )
		return follow ( * symbol, symbolptr );

	else if ( empty )
		return follow ( * empty, symbolptr );

	else if ( eps )
		return follow ( * eps, symbolptr );

	throw exception::CommonException ( "GlushkovTraversal::follow() - unknown RegExpElement node" );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExpAlternation & node, const regexp::UnboundedRegExpSymbol & symbolptr ) {
	for ( auto const & element : node.getElements ( ) )
		if ( pos ( * element, symbolptr ) )
			return follow ( * element, symbolptr );

	throw exception::CommonException ( "GlushkovTraversal::follow(Alt)" );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExpConcatenation & node, const regexp::UnboundedRegExpSymbol & symbolptr ) {
	std::set < regexp::UnboundedRegExpSymbol > ret, tmp, lastSet;

	for ( auto e = node.getElements ( ).begin ( ); e != node.getElements ( ).end ( ); e++ ) {
		if ( !pos ( * * e, symbolptr ) )
			continue;

		tmp = follow ( * * e, symbolptr );
		ret.insert ( tmp.begin ( ), tmp.end ( ) );

		lastSet = last ( * * e );

		if ( lastSet.find ( symbolptr ) != lastSet.end ( ) )
			for ( auto f = next ( e ); f != node.getElements ( ).end ( ); f++ ) {
				tmp = first ( * * f );
				ret.insert ( tmp.begin ( ), tmp.end ( ) );

				if ( !regexp::properties::RegExpEpsilon::languageContainsEpsilon ( * * f ) )
					break;
			}

	}

	return ret;
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExpIteration & node, const regexp::UnboundedRegExpSymbol & symbolptr ) {
	std::set < regexp::UnboundedRegExpSymbol > ret = follow ( node.getElement ( ), symbolptr );
	std::set < regexp::UnboundedRegExpSymbol > lastSet = last ( node.getElement ( ) );

	if ( lastSet.find ( symbolptr ) != lastSet.end ( ) ) {
		std::set < regexp::UnboundedRegExpSymbol > firstSet = first ( node.getElement ( ) );
		ret.insert ( firstSet.begin ( ), firstSet.end ( ) );
	}

	return ret;
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExpSymbol & /* node */, const regexp::UnboundedRegExpSymbol & /* symbolptr */ ) {
	return std::set < regexp::UnboundedRegExpSymbol > ( );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExpEmpty & /* node */, const regexp::UnboundedRegExpSymbol & /* symbolptr */ ) {
	return std::set < regexp::UnboundedRegExpSymbol > ( );
}

std::set < regexp::UnboundedRegExpSymbol > GlushkovTraversal::follow ( const regexp::UnboundedRegExpEpsilon & /* node */, const regexp::UnboundedRegExpSymbol & /* symbolptr */ ) {
	return std::set < regexp::UnboundedRegExpSymbol > ( );
}

// ----------------------------------------------------------------------------

bool GlushkovTraversal::pos ( const regexp::UnboundedRegExpElement & node, const regexp::UnboundedRegExpSymbol & symbolptr ) {
	const regexp::UnboundedRegExpAlternation * alternation = dynamic_cast < const regexp::UnboundedRegExpAlternation * > ( & node );
	const regexp::UnboundedRegExpConcatenation * concatenation = dynamic_cast < const regexp::UnboundedRegExpConcatenation * > ( & node );
	const regexp::UnboundedRegExpIteration * iteration = dynamic_cast < const regexp::UnboundedRegExpIteration * > ( & node );
	const regexp::UnboundedRegExpSymbol * symbol = dynamic_cast < const regexp::UnboundedRegExpSymbol * > ( & node );
	const regexp::UnboundedRegExpEmpty * empty = dynamic_cast < const regexp::UnboundedRegExpEmpty * > ( & node );
	const regexp::UnboundedRegExpEpsilon * eps = dynamic_cast < const regexp::UnboundedRegExpEpsilon * > ( & node );

	if ( alternation )
		return pos ( * alternation, symbolptr );

	else if ( concatenation )
		return pos ( * concatenation, symbolptr );

	else if ( iteration )
		return pos ( * iteration, symbolptr );

	else if ( symbol )
		return pos ( * symbol, symbolptr );

	else if ( empty )
		return pos ( * empty, symbolptr );

	else if ( eps )
		return pos ( * eps, symbolptr );

	throw exception::CommonException ( "GlushkovTraversal::pos() - unknown RegExpElement node" );
}

bool GlushkovTraversal::pos ( const regexp::UnboundedRegExpAlternation & node, const regexp::UnboundedRegExpSymbol & symbol ) {
	for ( const auto & element : node.getElements ( ) )
		if ( pos ( * element, symbol ) )
			return true;

	return false;
}

bool GlushkovTraversal::pos ( const regexp::UnboundedRegExpConcatenation & node, const regexp::UnboundedRegExpSymbol & symbol ) {
	for ( const auto & element : node.getElements ( ) )
		if ( pos ( * element, symbol ) )
			return true;

	return false;
}

bool GlushkovTraversal::pos ( const regexp::UnboundedRegExpIteration & node, const regexp::UnboundedRegExpSymbol & symbol ) {
	return pos ( node.getElement ( ), symbol );
}

bool GlushkovTraversal::pos ( const regexp::UnboundedRegExpSymbol & node, const regexp::UnboundedRegExpSymbol & symbol ) {
	return symbol == node;
}

bool GlushkovTraversal::pos ( const regexp::UnboundedRegExpEmpty & /* node */, const regexp::UnboundedRegExpSymbol & /* symbol */ ) {
	return false;
}

bool GlushkovTraversal::pos ( const regexp::UnboundedRegExpEpsilon & /* node */, const regexp::UnboundedRegExpSymbol & /* symbol */ ) {
	return false;
}

// ----------------------------------------------------------------------------

UnboundedRegExp GlushkovTraversal::indexate ( const regexp::UnboundedRegExp & re ) {
	int i = 1;

	return UnboundedRegExp ( indexate ( re.getRegExp ( ), i ) );
}

std::rvalue_ref < UnboundedRegExpElement > GlushkovTraversal::indexate ( const regexp::UnboundedRegExpElement & node, int & i ) {
	const regexp::UnboundedRegExpAlternation * alternation = dynamic_cast < const regexp::UnboundedRegExpAlternation * > ( & node );
	const regexp::UnboundedRegExpConcatenation * concatenation = dynamic_cast < const regexp::UnboundedRegExpConcatenation * > ( & node );
	const regexp::UnboundedRegExpIteration * iteration = dynamic_cast < const regexp::UnboundedRegExpIteration * > ( & node );
	const regexp::UnboundedRegExpSymbol * symbol = dynamic_cast < const regexp::UnboundedRegExpSymbol * > ( & node );
	const regexp::UnboundedRegExpEmpty * empty = dynamic_cast < const regexp::UnboundedRegExpEmpty * > ( & node );
	const regexp::UnboundedRegExpEpsilon * eps = dynamic_cast < const regexp::UnboundedRegExpEpsilon * > ( & node );

	if ( symbol ) {
		return std::rvalue_ref < UnboundedRegExpElement > ( new UnboundedRegExpSymbol ( alphabet::Symbol ( alphabet::SymbolPairSymbol ( std::make_pair ( symbol->getSymbol ( ), alphabet::symbolFrom ( i++ ) ) ) ) ) );
	} else if ( alternation ) {
		UnboundedRegExpAlternation * alt = new UnboundedRegExpAlternation();

		for ( const auto & element : alternation->getElements ( ) )
			alt->appendElement ( indexate ( * element, i ) );

		return std::rvalue_ref < UnboundedRegExpElement > ( alt );
	} else if ( concatenation ) {
		UnboundedRegExpConcatenation * con = new UnboundedRegExpConcatenation();

		for ( const auto & element : concatenation->getElements ( ) )
			con->appendElement ( indexate ( * element, i ) );

		return std::rvalue_ref < UnboundedRegExpElement > ( con );
	} else if ( iteration ) {
		return std::rvalue_ref < UnboundedRegExpElement > ( new UnboundedRegExpIteration ( indexate ( iteration->getElement ( ), i ) ) );
	} else if ( empty ) {
		return std::rvalue_ref < UnboundedRegExpElement > ( new UnboundedRegExpEmpty ( ) );
	} else if ( eps ) {
		return std::rvalue_ref < UnboundedRegExpElement > ( new UnboundedRegExpEpsilon ( ) );
	} else {
		throw exception::CommonException ( "GlushkovTraversal::getSymbols() - unknown RegExpElement node" );
	}
}

} /* namespace conversions */
