/*
 * RegExpOptimize.cpp
 *
 *  Created on: 20. 1. 2014
 *	  Author: Jan Travnicek
 */

namespace regexp {

namespace simplify {

FormalRegExpElement* RegExpOptimize::optimize( FormalRegExpElement const * const & node ) const
{
	FormalRegExpElement* elem = node->clone();

	// optimize while you can
	while(    A1( elem ) || A2( elem ) || A3( elem ) || A4( elem ) || A10( elem ) || V2( elem ) || V5( elem ) || V6( elem ) || X1( elem )
	       || A5( elem ) || A6( elem ) || A7( elem ) || A8( elem ) || A9( elem ) || V8( elem ) //|| V9( elem )
	       || A11( elem ) || V1( elem ) || V3( elem ) || V4( elem ) || V10( elem ) || S(elem) );

	return elem;
}

bool RegExpOptimize::S( FormalRegExpElement * & node ) const
{
	bool optimized = false;
	FormalRegExpAlternation * alternation = dynamic_cast<FormalRegExpAlternation*>( node );
	if( alternation ) {
		FormalRegExpElement * tmp = optimize ( alternation->left.get() );
		if(tmp != alternation->left.get()) {
			optimized = true;
			alternation->left = std::smart_ptr < FormalRegExpElement > ( tmp );
		}

		tmp = optimize ( alternation->right.get() );
		if(tmp != alternation->right.get()) {
			optimized = true;
			alternation->right = std::smart_ptr < FormalRegExpElement > ( tmp );
		}

		return optimized;
	}

	FormalRegExpConcatenation * concatenation = dynamic_cast<FormalRegExpConcatenation*>( node );
	if( concatenation ) {
		FormalRegExpElement* tmp = optimize ( concatenation->left.get() );
		if(tmp != concatenation->left.get()) {
			optimized = true;
			concatenation->left = std::smart_ptr < FormalRegExpElement > ( tmp );
		}

		tmp = optimize ( concatenation->right.get());
		if(tmp != concatenation->right.get()) {
			optimized = true;
			concatenation->right = std::smart_ptr < FormalRegExpElement > ( tmp );
		}

		return optimized;
	}

	FormalRegExpIteration * iteration = dynamic_cast<FormalRegExpIteration*>( node );
	if( iteration ) {
		FormalRegExpElement* tmp = optimize ( iteration->element.get() );

		if(tmp != iteration->element.get()) {
			optimized = true;
			iteration->element = std::smart_ptr < FormalRegExpElement > ( tmp );
		}
		return iteration;
	}

	return optimized;
}


/**
  * optimization A1: ( x + y ) + z = x + ( y + z )
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A1( FormalRegExpElement * & n ) const
{
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpAlternation * leftAlt = dynamic_cast<FormalRegExpAlternation *>( node->left.get() );

	if( leftAlt ) {
		leftAlt = static_cast<FormalRegExpAlternation *>( node->left.release() );
		std::smart_ptr < FormalRegExpElement > x = std::move ( leftAlt->left );
		std::smart_ptr < FormalRegExpElement > y = std::move ( leftAlt->right );
		std::smart_ptr < FormalRegExpElement > z = std::move ( node->right );

		leftAlt->left = std::move ( y );
		leftAlt->right = std::move ( z );
		node->left = std::move ( x );
		node->right = std::smart_ptr < FormalRegExpElement > ( leftAlt );

		return true;
	}

	return false;
}

/**
  * optimization A2: x + y = y + x (sort)
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A2( FormalRegExpElement * & n ) const
{
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpAlternation * rightAlt = dynamic_cast<FormalRegExpAlternation *>( node->right.get() );

	if( rightAlt ) {
		FormalRegExpElement * x = node->left.get();
		FormalRegExpElement * y = rightAlt->left.get();

		if(*x > *y) {
			std::swap ( node->left, rightAlt->left );
		} else {
			return false;
		}
	}

	return false;
}

/**
  * optimization A3: x + \0 = x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A3( FormalRegExpElement * & n ) const
{
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	// input can be \0 + \0, so at least one element must be preserved

	FormalRegExpEmpty * rightEmp = dynamic_cast<FormalRegExpEmpty *>( node->right.get() );
	if( rightEmp ) {
		rightEmp = nullptr;
		n = node->left.release();
		delete node;
		return true;
	}

	FormalRegExpEmpty * leftEmp = dynamic_cast<FormalRegExpEmpty *>( node->left.release() );
	if( leftEmp ) {
		leftEmp = nullptr;
		n = node->right.release();
		delete node;
		return true;
	}

	return false;
}

/**
  * optimization A4: x + x = x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A4( FormalRegExpElement * & n ) const
{
	/*
	 * two ways of implementing this opitimization:
	 * - sort and call std::unique ( O(n lg n) + O(n) ), but it also sorts...
	 * - check every element against other ( O(n*n) )
	 *
	 * As we always sort in optimization, we can use the first version, but A4 must be __always__ called __after__ A2
	 */

	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	if( *node->left == *node->right ) {
		delete node->left.release();
		n = node->right.release();
		delete node;
		return true;
	}

	return false;
}

/**
  * optimization A5: x.(y.z) = (x.y).z = x.y.z
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A5( FormalRegExpElement * & n ) const
{
	FormalRegExpConcatenation * node = dynamic_cast<FormalRegExpConcatenation *>( n );
	if( ! node ) return false;

	FormalRegExpConcatenation * leftCon = dynamic_cast<FormalRegExpConcatenation *>( node->left.get() );

	if( leftCon ) {
		leftCon = static_cast<FormalRegExpConcatenation *>( node->left.release() );
		std::smart_ptr < FormalRegExpElement > x = std::move ( leftCon->left );
		std::smart_ptr < FormalRegExpElement > y = std::move ( leftCon->right );
		std::smart_ptr < FormalRegExpElement > z = std::move ( node->right );

		leftCon->left = std::move ( y );
		leftCon->right = std::move ( z );
		node->left = std::move ( x );
		node->right = std::smart_ptr < FormalRegExpElement > ( leftCon );

		return true;
	}

	return false;
}

/**
  * optimization A6: \e.x = x.\e = x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A6( FormalRegExpElement * & n ) const
{
	FormalRegExpConcatenation * node = dynamic_cast<FormalRegExpConcatenation *>( n );
	if( ! node ) return false;

	// input can be \e + \e, so at least one element must be preserved

	FormalRegExpEpsilon * rightEmp = dynamic_cast<FormalRegExpEpsilon *>( node->right.get() );
	if( rightEmp ) {
		rightEmp = nullptr;
		n = node->left.release();
		delete node;
		return true;
	}

	FormalRegExpEpsilon * leftEmp = dynamic_cast<FormalRegExpEpsilon *>( node->left.get() );
	if( leftEmp ) {
		leftEmp = nullptr;
		n = node->right.release();
		delete node;
		return true;
	}

	return false;
}

/**
  * optimization A7: \0.x = x.\0 = \0
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A7( FormalRegExpElement * & n ) const
{
	FormalRegExpConcatenation * node = dynamic_cast<FormalRegExpConcatenation *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpEmpty *>( node->right.get() ) || dynamic_cast<FormalRegExpEmpty *>( node->left.get() ) ) {
		delete node;
		n = new FormalRegExpEmpty { };
		return true;
	}

	return false;
}

/**
  * optimization A8: x.(y+z) = x.y + x.z
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A8( FormalRegExpElement * & /* n */) const
{
	return false; //TODO
}

/**
  * optimization A9: (x+y).z = x.z + y.z
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A9( FormalRegExpElement * & /* n */) const
{
	return false; //TODO
}

/**
  * optimization A10: x* = \e + x*x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A10( FormalRegExpElement * & n ) const
{
	/*
	 * problem:
	 * - \e + x*x = x*
	 * - but if we do not have the eps, but we do have iteration, then \e \in h(iter), therefore \e in h(node).
	 */

	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpEpsilon * leftEps = dynamic_cast<FormalRegExpEpsilon *>( node->left.get() );
	if( leftEps ) {
		FormalRegExpConcatenation * rightCon = dynamic_cast<FormalRegExpConcatenation *>( node->right.get() );
		if( ! rightCon ) return false;

		FormalRegExpIteration * rightLeftIte = dynamic_cast<FormalRegExpIteration *>( rightCon->left.get() );
		if( rightLeftIte ) {
			if(*rightLeftIte->element == *rightCon->right) {
				n = rightCon->left.release();
				delete node;
				return true;
			}
		}

		FormalRegExpIteration * rightRightIte = dynamic_cast<FormalRegExpIteration *>( rightCon->right.get() );
		if( rightRightIte ) {
			if(*rightLeftIte->element == *rightCon->left) {
				n = rightCon->right.release();
				delete node;
				return true;
			}
		}
	}

	FormalRegExpEpsilon * rightEps = dynamic_cast<FormalRegExpEpsilon *>( node->right.get() );
	if( rightEps ) {
		FormalRegExpConcatenation * leftCon = dynamic_cast<FormalRegExpConcatenation *>( node->left.get() );
		if( ! leftCon ) return false;

		FormalRegExpIteration * leftLeftIte = dynamic_cast<FormalRegExpIteration *>( leftCon->left.get() );
		if( leftLeftIte ) {
			if(*leftLeftIte->element == *leftCon->right) {
				n = leftCon->left.release();
				delete node;
				return true;
			}
		}

		FormalRegExpIteration * leftRightIte = dynamic_cast<FormalRegExpIteration *>( leftCon->right.get() );
		if( leftRightIte ) {
			if(*leftLeftIte->element == *leftCon->left) {
				n = leftCon->right.release();
				delete node;
				return true;
			}
		}
	}

	return false;
}

/**
  * optimization A11: x* = (\e + x)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A11( FormalRegExpElement * & n ) const
{
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	FormalRegExpAlternation * childAlt = dynamic_cast<FormalRegExpAlternation *>( node->element.get() );
	if( childAlt )
	{
		if(dynamic_cast<FormalRegExpEpsilon*>(childAlt->left.get())) {
			node->element = std::move ( childAlt->right );
			return true;
		}
		if(dynamic_cast<FormalRegExpEpsilon*>(childAlt->right.get())) {
			node->element = std::move ( childAlt->left );
			return true;
		}
	}

	return false;
}

/**
  * optimization V1: \0* = \e
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V1( FormalRegExpElement * & n ) const
{
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpEmpty*>( node->element.get() ) )
	{
		delete node;
		n = new FormalRegExpEpsilon( );
		return true;
	}
	return false;
}

/**
  * optimization V2: x* + x = x*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V2( FormalRegExpElement * & n ) const
{
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( node->left.get() );
	if( leftIte ) {
		if(*leftIte->element == *node->right) {
			n = node->left.release();
			delete node;
			return true;
		}
	}

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( node->right.get() );
	if( rightIte ) {
		if(*rightIte->element == *node->left) {
			n = node->right.release();
			delete node;
			return true;
		}
	}

	return false;
}

/**
  * optimization V3: x** = x*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V3( FormalRegExpElement * & n ) const
{
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	FormalRegExpIteration* childIter = dynamic_cast<FormalRegExpIteration*>( node->element.get() );
	if( childIter )
	{
		node->element = std::move ( childIter->element );
		return true;
	}

	return false;
}

/**
  * optimization V4: (x+y)* = (x*y*)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V4( FormalRegExpElement * & n ) const
{
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	FormalRegExpConcatenation * child = dynamic_cast<FormalRegExpConcatenation *>( node->element.get() );
	if( ! child ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( child->left.get() );
	if( ! leftIte ) return false;

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( child->right.get() );
	if( ! rightIte ) return false;

	n = new FormalRegExpIteration( FormalRegExpAlternation(std::move( * leftIte->element ), std::move( * rightIte->element)));

	delete node;
	return true;
}

/**
  * optimization V5: x*y = y + x*xy
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V5( FormalRegExpElement * & /* n */) const
{
	return false; //TODO
}

/**
  * optimization V6: x*y = y + xx*y
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V6( FormalRegExpElement * & /* n */) const
{
	return false; //TODO
}

/**
  * optimization V8: \e in h(x) => xx*=x*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V8( FormalRegExpElement * & /* n */) const
{
	return false; //TODO
}

/**
  * optimization V9: (xy)*x = x(yx)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V9( FormalRegExpElement * & /* n */) const
{
	return false; //TODO
}

/**
  * optimization V10: (x+y)* = (x*+y*)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V10( FormalRegExpElement * & n ) const
{
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( node->left.get() );
	if( ! leftIte ) return false;

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( node->right.get() );
	if( ! rightIte ) return false;

	n = new FormalRegExpConcatenation(std::move( * leftIte->element ), std::move( * rightIte->element ) );

	delete node;
	return true;
}

/**
  * optimization X1: a* + \e = a*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::X1( FormalRegExpElement * & n ) const
{
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( node->left.get() );
	if( leftIte ) {
		if(dynamic_cast<FormalRegExpEpsilon*>(node->right.get())) {
			n = node->left.release();
			delete node;
			return true;
		}
	}

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( node->right.get() );
	if( rightIte ) {
		if(dynamic_cast<FormalRegExpEpsilon*>(node->left.get())) {
			n = node->right.release();
			delete node;
			return true;
		}
	}

	return false;

}

} /* namespace simplify */

} /* namespace regexp */
