/*
 * REAL_TIME_HEIGHT_DETERMINISTIC_NPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_
#define REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_

#include <map>
#include <vector>
#include <variant>
#include "../AutomatonBase.h"
#include "../common/State.h"
#include "../../alphabet/Symbol.h"
#include "../../string/Epsilon.h"

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialStates;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class RealTimeHeightDeterministicNPDA : public AutomatonBase, public std::Components < RealTimeHeightDeterministicNPDA, alphabet::Symbol, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, automaton::State, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > {
protected:
	std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < std::pair < State, alphabet::Symbol > > > callTransitions;
	std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, alphabet::Symbol >, std::set < State > > returnTransitions;
	std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > localTransitions;

public:
	explicit RealTimeHeightDeterministicNPDA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, std::set < alphabet::Symbol > pushdownStoreSymbol, std::set < State > initialStates, alphabet::Symbol bottomOfTheStackSymbol, std::set < automaton::State > finalStates );
	explicit RealTimeHeightDeterministicNPDA ( alphabet::Symbol bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < State > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( State state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < State > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const State & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < State > & getInitialStates ( ) const {
		return accessComponent < InitialStates > ( ).get ( );
	}

	bool addInitialState ( State state ) {
		return accessComponent < InitialStates > ( ).add ( std::move ( state ) );
	}

	void setInitialStates ( std::set < State > states ) {
		accessComponent < InitialStates > ( ).set ( std::move ( states ) );
	}

	void removeInitialState ( const State & state ) {
		accessComponent < InitialStates > ( ).remove ( state );
	}

	const std::set < State > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( State state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < State > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const State & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getPushdownStoreAlphabet ( ) const {
		return accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const alphabet::Symbol & getBottomOfTheStackSymbol ( ) const {
		return accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( alphabet::Symbol symbol ) {
		return accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( State current, std::variant < string::Epsilon, alphabet::Symbol > input, State next, alphabet::Symbol push );
	bool addCallTransition ( State current, State next, alphabet::Symbol push );
	bool addCallTransition ( State current, alphabet::Symbol input, State next, alphabet::Symbol push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( State current, std::variant < string::Epsilon, alphabet::Symbol > input, alphabet::Symbol pop, State next );
	bool addReturnTransition ( State current, alphabet::Symbol pop, State next );
	bool addReturnTransition ( State current, alphabet::Symbol input, alphabet::Symbol pop, State next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( State current, std::variant < string::Epsilon, alphabet::Symbol > input, State next );
	bool addLocalTransition ( State current, State next );
	bool addLocalTransition ( State current, alphabet::Symbol input, State next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const State & current, const std::variant < string::Epsilon, alphabet::Symbol > & input, const State & next, const alphabet::Symbol & push );
	bool removeCallTransition ( const State & current, const State & next, const alphabet::Symbol & push );
	bool removeCallTransition ( const State & current, const alphabet::Symbol & input, const State & next, const alphabet::Symbol & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const State & current, const std::variant < string::Epsilon, alphabet::Symbol > & input, const alphabet::Symbol & pop, const State & next );
	bool removeReturnTransition ( const State & current, const alphabet::Symbol & pop, const State & next );
	bool removeReturnTransition ( const State & current, const alphabet::Symbol & input, const alphabet::Symbol & pop, const State & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const State & current, const std::variant < string::Epsilon, alphabet::Symbol > & input, const State & next );
	bool removeLocalTransition ( const State & current, const State & next );
	bool removeLocalTransition ( const State & current, const alphabet::Symbol & input, const State & next );

	const std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < std::pair < State, alphabet::Symbol > > > & getCallTransitions ( ) const;

	const std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, alphabet::Symbol >, std::set < State > > & getReturnTransitions ( ) const;

	const std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RealTimeHeightDeterministicNPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RealTimeHeightDeterministicNPDA";

		return xmlTagName;
	}

	static RealTimeHeightDeterministicNPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, RealTimeHeightDeterministicNPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

#endif /* REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_ */
