/*
 * GrammarToXMLComposer.cpp
 *
 *  Created on: Nov 11, 2013
 *      Author: Jan Travnicek
 */

#include "GrammarToXMLComposer.h"

#include <XmlApi.hpp>
#include "../../alphabet/Symbol.h"

namespace grammar {

void GrammarToXMLComposer::composeNonterminalAlphabet(std::deque<sax::Token>& out, const std::set<alphabet::Symbol>& symbols) {
	out.emplace_back("nonterminalAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	}
	out.emplace_back("nonterminalAlphabet", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeTerminalAlphabet(std::deque<sax::Token>& out, const std::set<alphabet::Symbol>& symbols) {
	out.emplace_back("terminalAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	}
	out.emplace_back("terminalAlphabet", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeInitialSymbol(std::deque<sax::Token>& out, const alphabet::Symbol& symbol) {
	out.emplace_back("initialSymbol", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	out.emplace_back("initialSymbol", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeGeneratesEpsilon(std::deque<sax::Token>& out, bool generatesEpsilon) {
	out.emplace_back("generatesEpsilon", sax::Token::TokenType::START_ELEMENT);
	if(generatesEpsilon) {
		out.emplace_back("true", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("true", sax::Token::TokenType::END_ELEMENT);
	} else {
		out.emplace_back("false", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("false", sax::Token::TokenType::END_ELEMENT);
	}
	out.emplace_back("generatesEpsilon", sax::Token::TokenType::END_ELEMENT);
}


void GrammarToXMLComposer::composeRuleLContext(std::deque<sax::Token>& out, const std::vector<alphabet::Symbol>& symbols) {
	out.emplace_back("lContext", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	}
	out.emplace_back("lContext", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleLHS(std::deque<sax::Token>& out, const std::vector<alphabet::Symbol>& symbols) {
	out.emplace_back("lhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	}
	out.emplace_back("lhs", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleRContext(std::deque<sax::Token>& out, const std::vector<alphabet::Symbol>& symbols) {
	out.emplace_back("rContext", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	}
	out.emplace_back("rContext", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleSingleSymbolLHS(std::deque<sax::Token>& out, const alphabet::Symbol& symbol) {
	out.emplace_back("lhs", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	out.emplace_back("lhs", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleRHS(std::deque<sax::Token>& out, const std::vector<alphabet::Symbol>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleOneOrTwoSymbolsRHS(std::deque<sax::Token>& out, const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.is<alphabet::Symbol>()) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbols.get<alphabet::Symbol>());
	} else {
		const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = symbols.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
		alib::xmlApi<alphabet::Symbol>::compose(out, rhs.first);
		alib::xmlApi<alphabet::Symbol>::compose(out, rhs.second);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleGNFRHS(std::deque<sax::Token>& out, const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);

	alib::xmlApi<alphabet::Symbol>::compose(out, symbols.first);

	for (const auto& symbol : symbols.second) {
		alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
	}

	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleLGRHS(std::deque<sax::Token>& out, const std::variant<std::vector<alphabet::Symbol>, std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.is<std::vector<alphabet::Symbol>>()) {
		const std::vector<alphabet::Symbol>& rhs = symbols.get<std::vector<alphabet::Symbol>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else for (const auto& symbol : symbols.get<std::vector<alphabet::Symbol>>()) {
			alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
		}
	} else {
		const std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>& rhs = symbols.get<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>>();
		for (const auto& symbol : std::get<0>(rhs)) {
			alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
		}
		alib::xmlApi<alphabet::Symbol>::compose(out, std::get<1>(rhs));
		for (const auto& symbol : std::get<2>(rhs)) {
			alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
		}
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleLeftLGRHS(std::deque<sax::Token>& out, const std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.is<std::vector<alphabet::Symbol>>()) {
		const std::vector<alphabet::Symbol>& rhs = symbols.get<std::vector<alphabet::Symbol>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else for (const auto& symbol : symbols.get<std::vector<alphabet::Symbol>>()) {
			alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
		}
	} else {
		const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>& rhs = symbols.get<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>>();
		alib::xmlApi<alphabet::Symbol>::compose(out, rhs.first);
		for (const auto& symbol : rhs.second) {
			alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
		}
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

void GrammarToXMLComposer::composeRuleRightLGRHS(std::deque<sax::Token>& out, const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.is<std::vector<alphabet::Symbol>>()) {
		const std::vector<alphabet::Symbol>& rhs = symbols.get<std::vector<alphabet::Symbol>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else {
			for (const auto& symbol : symbols.get<std::vector<alphabet::Symbol>>()) {
				alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
			}
		}
	} else {
		const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rhs = symbols.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();
		for (const auto& symbol : rhs.first) {
			alib::xmlApi<alphabet::Symbol>::compose(out, symbol);
		}
		alib::xmlApi<alphabet::Symbol>::compose(out, rhs.second);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

