/*
 * AdjacencyListUndirectedGraph.h
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#ifndef ADJACENCY_LIST_UNDIRECTED_GRAPH_H_
#define ADJACENCY_LIST_UNDIRECTED_GRAPH_H_

#include <list>
#include <unordered_map>

#include "UndirectedGraph.h"

namespace graph {

class DirectedGraph;

/*
 *  AdjacencyList representation
 *
 *  unordered_map:
 *    [ n1 ] -> list [ n2, n2, n3, n4 ]
 *    [ n2 ] -> list [ n1, n1, n4 ]      < multi-edge
 *    [ n3 ] -> list [ n1, n4 ]
 *    [ n4 ] -> list [ n1, n2, n3 ]
 *
 *  edges:
 *    n1 <-> n2, n1 <-> n2, n1 <-> n3, n1 <-> n4
 *    n2 <-> n4
 *    n3 <-> n4
 *
 *  undirected edge n1 <-> n2 is represented as
 *  directed edge n1 -> n2 and directed edge n2 -> n1
 */

class AdjacencyListUndirectedGraph : public UndirectedGraph {
	std::unordered_map< Node, std::list< Node > > adj;
	std::set< UndirectedEdge > edges;

public:
	AdjacencyListUndirectedGraph();

	/// constructs an undirected adjacency list from any representation of undirected graph
	AdjacencyListUndirectedGraph( const UndirectedGraph & graph );
	/// converts directed graph into undirected by removing edge direction
	AdjacencyListUndirectedGraph( const DirectedGraph & dgraph );

	virtual GraphBase * clone() const override;
	virtual GraphBase * plunder() && override;

	virtual REPRESENTATION getRepresentation() const override;
	virtual std::string className() const override;

	virtual std::set< Node > getNodes() const override;
	virtual std::set< UndirectedEdge > getEdges() const override;

	virtual size_t nodeCount() const override;
	virtual size_t edgeCount() const override;

	virtual bool hasNode( const Node & node ) const override;
	virtual bool hasEdge( const UndirectedEdge & edge ) const override;

	virtual bool hasEdge( const Node & first, const Node & second ) const override;
	virtual std::set< UndirectedEdge > findEdges( const Node & first, const Node & second ) const override;

	virtual std::set< Node > neighbors( const Node & node ) const override;
	virtual std::set< UndirectedEdge > neighborEdges( const Node & node ) const override;

	virtual bool addNode( const Node & node ) override;
	virtual bool addNode( const Node & node, int value ) override;
	virtual bool removeNode( const Node & node ) override;

	virtual bool addEdge( const UndirectedEdge &edge ) override;
	virtual bool addEdge( const UndirectedEdge & edge, int value ) override;
	virtual bool removeEdge( const UndirectedEdge &edge ) override;

	static AdjacencyListUndirectedGraph parse( std::deque< sax::Token >::iterator & input );
	virtual void compose( std::deque< sax::Token > & out ) const override;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "AdjacencyListUndirectedGraph";

		return xmlTagName;
	}
};

} // namespace graph

#endif // ADJACENCY_LIST_DIRECTED_GRAPH_H_
