/*
 * FormalRegExpAlternation.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "FormalRegExpAlternation.h"
#include <exception/CommonException.h>
#include "../unbounded/UnboundedRegExpAlternation.h"
#include <sstream>

namespace regexp {

FormalRegExpAlternation::FormalRegExpAlternation ( FormalRegExpElement && leftElement, FormalRegExpElement && rightElement ) : left ( nullptr ), right ( nullptr ) {
	setLeftElement ( std::move ( leftElement ) );
	setRightElement ( std::move ( rightElement ) );
}

FormalRegExpAlternation::FormalRegExpAlternation ( const FormalRegExpElement & left, const FormalRegExpElement & right ) : FormalRegExpAlternation ( std::move_copy ( left ), std::move_copy ( right ) ) {
}

FormalRegExpAlternation::FormalRegExpAlternation ( const FormalRegExpAlternation & other ) : left ( other.left ), right ( other.right ) {
}

FormalRegExpAlternation::FormalRegExpAlternation ( FormalRegExpAlternation && other ) noexcept : left ( std::move ( other.left ) ), right ( std::move ( other.right ) ) {
	left->attachRegExp ( NULL );
	right->attachRegExp ( NULL );
}

FormalRegExpAlternation & FormalRegExpAlternation::operator =( const FormalRegExpAlternation & other ) {
	if ( this == & other )
		return * this;

	* this = FormalRegExpAlternation ( other );

	return * this;
}

FormalRegExpAlternation & FormalRegExpAlternation::operator =( FormalRegExpAlternation && other ) noexcept {
	std::swap ( this->left, other.left );
	std::swap ( this->right, other.right );

	this->left->attachRegExp ( this->parentRegExp );
	this->right->attachRegExp ( this->parentRegExp );

	return * this;
}

FormalRegExpAlternation::~FormalRegExpAlternation ( ) noexcept {
}

const FormalRegExpElement & FormalRegExpAlternation::getLeftElement ( ) const {
	return * left;
}

const FormalRegExpElement & FormalRegExpAlternation::getRightElement ( ) const {
	return * right;
}

FormalRegExpElement & FormalRegExpAlternation::getLeftElement ( ) {
	return * left;
}

FormalRegExpElement & FormalRegExpAlternation::getRightElement ( ) {
	return * right;
}

void FormalRegExpAlternation::setLeftElement ( FormalRegExpElement && element ) {
	this->left = std::smart_ptr < FormalRegExpElement > ( std::move ( element ).plunder ( ) );

	if ( this->parentRegExp && !left->attachRegExp ( this->parentRegExp ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void FormalRegExpAlternation::setLeftElement ( const FormalRegExpElement & element ) {
	setLeftElement ( std::move_copy ( element ) );
}

void FormalRegExpAlternation::setRightElement ( FormalRegExpElement && element ) {
	this->right = std::smart_ptr < FormalRegExpElement > ( std::move ( element ).plunder ( ) );

	if ( this->parentRegExp && !right->attachRegExp ( this->parentRegExp ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void FormalRegExpAlternation::setRightElement ( const FormalRegExpElement & element ) {
	setRightElement ( std::move_copy ( element ) );
}

FormalRegExpElement * FormalRegExpAlternation::clone ( ) const {
	return new FormalRegExpAlternation ( * this );
}

FormalRegExpElement * FormalRegExpAlternation::plunder ( ) && {
	return new FormalRegExpAlternation ( std::move ( * this ) );
}

std::smart_ptr < UnboundedRegExpElement > FormalRegExpAlternation::asUnbounded ( ) const {
	UnboundedRegExpAlternation * res = new UnboundedRegExpAlternation ( );

	res->appendElement ( std::move ( * left->asUnbounded ( ) ) );
	res->appendElement ( std::move ( * right->asUnbounded ( ) ) );

	return std::smart_ptr < UnboundedRegExpElement > ( res );
}

int FormalRegExpAlternation::compare ( const FormalRegExpAlternation & other ) const {
	int res = left->compare ( * other.left );

	if ( res == 0 ) res = right->compare ( * other.right );

	return res;
}

void FormalRegExpAlternation::operator >>( std::ostream & out ) const {
	out << "(FormalRegExpAlternation";
	out << " " << * left;
	out << " " << * right;
	out << ")";
}

bool FormalRegExpAlternation::testSymbol ( const alphabet::Symbol & symbol ) const {
	if ( left->testSymbol ( symbol ) ) return true;

	if ( right->testSymbol ( symbol ) ) return true;

	return false;
}

bool FormalRegExpAlternation::attachRegExp ( const FormalRegExp * regexp ) {
	if ( this->parentRegExp == regexp ) return true;

	this->parentRegExp = regexp;
	return left->attachRegExp ( regexp ) && right->attachRegExp ( regexp );
}

void FormalRegExpAlternation::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	left->computeMinimalAlphabet ( alphabet );
	right->computeMinimalAlphabet ( alphabet );
}

FormalRegExpAlternation::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */
