/*
 * FormalRegExpConcatenation.cpp
 *
 *  Created on: Nov 27, 2013
 *      Author: Martin Zak
 */

#include "FormalRegExpConcatenation.h"
#include <exception/CommonException.h>
#include "../unbounded/UnboundedRegExpConcatenation.h"

namespace regexp {

FormalRegExpConcatenation::FormalRegExpConcatenation ( FormalRegExpElement && leftElement, FormalRegExpElement && rightElement ) : left ( nullptr ), right ( nullptr ) {
	setLeftElement ( std::move ( leftElement ) );
	setRightElement ( std::move ( rightElement ) );
}

FormalRegExpConcatenation::FormalRegExpConcatenation ( const FormalRegExpElement & left, const FormalRegExpElement & right ) : FormalRegExpConcatenation ( std::move_copy ( left ), std::move_copy ( right ) ) {
}

FormalRegExpConcatenation::FormalRegExpConcatenation ( const FormalRegExpConcatenation & other ) : left ( other.left ), right ( other.right ) {
}

FormalRegExpConcatenation::FormalRegExpConcatenation ( FormalRegExpConcatenation && other ) noexcept : left ( std::move ( other.left ) ), right ( std::move ( other.right ) ) {
	left->attachRegExp ( NULL );
	right->attachRegExp ( NULL );
}

FormalRegExpConcatenation & FormalRegExpConcatenation::operator =( const FormalRegExpConcatenation & other ) {
	if ( this == & other )
		return * this;

	* this = FormalRegExpConcatenation ( other );

	return * this;
}

FormalRegExpConcatenation & FormalRegExpConcatenation::operator =( FormalRegExpConcatenation && other ) noexcept {
	std::swap ( this->left, other.left );
	std::swap ( this->right, other.right );

	this->left->attachRegExp ( this->parentRegExp );
	this->right->attachRegExp ( this->parentRegExp );

	return * this;
}

FormalRegExpConcatenation::~FormalRegExpConcatenation ( ) noexcept {
}

const FormalRegExpElement & FormalRegExpConcatenation::getLeftElement ( ) const {
	return * left;
}

const FormalRegExpElement & FormalRegExpConcatenation::getRightElement ( ) const {
	return * right;
}

FormalRegExpElement & FormalRegExpConcatenation::getLeftElement ( ) {
	return * left;
}

FormalRegExpElement & FormalRegExpConcatenation::getRightElement ( ) {
	return * right;
}

void FormalRegExpConcatenation::setLeftElement ( FormalRegExpElement && element ) {
	this->left = std::smart_ptr < FormalRegExpElement > ( std::move ( element ).plunder ( ) );

	if ( this->parentRegExp && !left->attachRegExp ( this->parentRegExp ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void FormalRegExpConcatenation::setLeftElement ( const FormalRegExpElement & element ) {
	setLeftElement ( std::move_copy ( element ) );
}

void FormalRegExpConcatenation::setRightElement ( FormalRegExpElement && element ) {
	this->right = std::smart_ptr < FormalRegExpElement > ( std::move ( element ).plunder ( ) );

	if ( this->parentRegExp && !right->attachRegExp ( this->parentRegExp ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void FormalRegExpConcatenation::setRightElement ( const FormalRegExpElement & element ) {
	setRightElement ( std::move_copy ( element ) );
}

FormalRegExpElement * FormalRegExpConcatenation::clone ( ) const {
	return new FormalRegExpConcatenation ( * this );
}

std::smart_ptr < UnboundedRegExpElement > FormalRegExpConcatenation::asUnbounded ( ) const {
	UnboundedRegExpConcatenation * res = new UnboundedRegExpConcatenation ( );

	res->appendElement ( * left->asUnbounded ( ) );
	res->appendElement ( * right->asUnbounded ( ) );

	return std::smart_ptr < UnboundedRegExpElement > ( res );
}

FormalRegExpElement * FormalRegExpConcatenation::plunder ( ) && {
	return new FormalRegExpConcatenation ( std::move ( * this ) );
}

int FormalRegExpConcatenation::compare ( const FormalRegExpConcatenation & other ) const {
	int res = left->compare ( * other.left );

	if ( res == 0 ) res = right->compare ( * other.right );

	return res;
}

void FormalRegExpConcatenation::operator >>( std::ostream & out ) const {
	out << "(FormalRegExpConcatenation";
	out << " " << * left;
	out << " " << * right;
	out << ")";
}

bool FormalRegExpConcatenation::testSymbol ( const alphabet::Symbol & symbol ) const {
	return left->testSymbol ( symbol ) && right->testSymbol ( symbol );
}

bool FormalRegExpConcatenation::attachRegExp ( const FormalRegExp * regexp ) {
	if ( this->parentRegExp == regexp ) return true;

	this->parentRegExp = regexp;
	return left->attachRegExp ( regexp ) && right->attachRegExp ( regexp );
}

void FormalRegExpConcatenation::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	left->computeMinimalAlphabet ( alphabet );
	right->computeMinimalAlphabet ( alphabet );
}

FormalRegExpConcatenation::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */
