/*
 * UnboundedRegExp.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "UnboundedRegExp.h"
#include <exception/CommonException.h>
#include "UnboundedRegExpEmpty.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../formal/FormalRegExp.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/RegExpFromXMLParser.h"
#include "../common/RegExpToXMLComposer.h"
#include "../RegExp.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace regexp {

UnboundedRegExp::UnboundedRegExp ( std::set < alphabet::Symbol > alphabet, UnboundedRegExpElement && regExp ) : std::Components < UnboundedRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), regExp ( NULL ) {
	setRegExp ( std::move ( regExp ) );
}

UnboundedRegExp::UnboundedRegExp ( std::set < alphabet::Symbol > alphabet, const UnboundedRegExpElement & regExp ) : UnboundedRegExp ( std::move ( alphabet ), std::move_copy ( regExp ) ) {
}

UnboundedRegExp::UnboundedRegExp ( ) : UnboundedRegExp ( std::set < alphabet::Symbol > ( ), UnboundedRegExpEmpty ( ) ) {
}

UnboundedRegExp::UnboundedRegExp ( UnboundedRegExpElement && regExp ) : UnboundedRegExp ( regExp.computeMinimalAlphabet ( ), std::move ( regExp ) ) {
}

UnboundedRegExp::UnboundedRegExp ( const UnboundedRegExpElement & regExp ) : UnboundedRegExp ( regExp.computeMinimalAlphabet ( ), std::move_copy ( regExp ) ) {
}

UnboundedRegExp::UnboundedRegExp ( const FormalRegExp & other ) : UnboundedRegExp ( other.getAlphabet ( ), * other.getRegExp ( ).asUnbounded ( ) ) {
}

UnboundedRegExp::UnboundedRegExp ( const UnboundedRegExp & other ) : std::Components < UnboundedRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( other.getAlphabet ( ) ), std::tuple < > ( ) ), regExp ( other.regExp ) {
	this->regExp->attachRegExp ( this );
}

UnboundedRegExp::UnboundedRegExp ( UnboundedRegExp && other ) noexcept : std::Components < UnboundedRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( other.accessComponent < GeneralAlphabet > ( ).get ( ) ) ), std::tuple < > ( ) ), regExp ( std::move ( other.regExp ) ) {
	this->regExp->attachRegExp ( this );
}

RegExpBase * UnboundedRegExp::clone ( ) const {
	return new UnboundedRegExp ( * this );
}

RegExpBase * UnboundedRegExp::plunder ( ) && {
	return new UnboundedRegExp ( std::move ( * this ) );
}

UnboundedRegExp & UnboundedRegExp::operator =( const UnboundedRegExp & other ) {
	if ( this == & other )
		return * this;

	* this = UnboundedRegExp ( other );

	return * this;
}

UnboundedRegExp & UnboundedRegExp::operator =( UnboundedRegExp && other ) noexcept {
	std::swap ( this->regExp, other.regExp );
	std::swap ( accessComponent < GeneralAlphabet > ( ).get ( ), other.accessComponent < GeneralAlphabet > ( ).get ( ) );

	return * this;
}

UnboundedRegExp::~UnboundedRegExp ( ) noexcept {
}

const UnboundedRegExpElement & UnboundedRegExp::getRegExp ( ) const {
	return * regExp;
}

UnboundedRegExpElement & UnboundedRegExp::getRegExp ( ) {
	return * regExp;
}

void UnboundedRegExp::setRegExp ( const UnboundedRegExpElement & regExp ) {
	setRegExp ( std::move_copy ( regExp ) );
}

void UnboundedRegExp::setRegExp ( UnboundedRegExpElement && param ) {
	this->regExp = std::smart_ptr < UnboundedRegExpElement > ( std::move ( param ).plunder ( ) );

	if ( !regExp->attachRegExp ( this ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void UnboundedRegExp::operator >>( std::ostream & out ) const {
	out << "(UnboundedRegExp " << * ( this->regExp ) << ")";
}

int UnboundedRegExp::compare ( const UnboundedRegExp & other ) const {
	auto first = std::tie ( regExp, getAlphabet ( ) );
	auto second = std::tie ( other.regExp, other.getAlphabet ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

UnboundedRegExp::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

UnboundedRegExp UnboundedRegExp::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnboundedRegExp::getXmlTagName() );

	std::set < alphabet::Symbol > alphabet = RegExpFromXMLParser::parseAlphabet ( input );

	std::rvalue_ref < UnboundedRegExpElement > element ( alib::xmlApi < regexp::UnboundedRegExpElement >::parse ( input ) );
	UnboundedRegExp regexp ( std::move ( alphabet ), std::move ( element ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnboundedRegExp::getXmlTagName() );

	return regexp;
}

void UnboundedRegExp::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnboundedRegExp::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	RegExpToXMLComposer::composeAlphabet ( out, this->getAlphabet ( ) );
	alib::xmlApi < UnboundedRegExpElement >::compose ( out, this->getRegExp ( ) );
	out.emplace_back ( UnboundedRegExp::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace regexp */

namespace std {

template < >
bool regexp::UnboundedRegExp::Component < regexp::UnboundedRegExp, alphabet::Symbol, regexp::GeneralAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	return static_cast < const regexp::UnboundedRegExp * > ( this )->getRegExp ( ).testSymbol ( symbol );
}

template < >
bool regexp::UnboundedRegExp::Component < regexp::UnboundedRegExp, alphabet::Symbol, regexp::GeneralAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void regexp::UnboundedRegExp::Component < regexp::UnboundedRegExp, alphabet::Symbol, regexp::GeneralAlphabet >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto unboundedRegExpParserRegister = xmlApi < regexp::RegExp >::ParserRegister < regexp::UnboundedRegExp > ( );
auto unboundedRegExpParserRegister2 = xmlApi < alib::Object >::ParserRegister < regexp::UnboundedRegExp > ( );

auto UnboundedRegExpFromFormalRegExp = castApi::CastRegister < regexp::UnboundedRegExp, regexp::FormalRegExp > ( );
auto UnboundedRegExpCastBinder = castApi::CastPoolStringBinder < regexp::UnboundedRegExp > ( regexp::UnboundedRegExp::getXmlTagName() );

} /* namespace alib */
