/*
 * UnboundedRegExpAlternation.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "UnboundedRegExpAlternation.h"
#include <exception/CommonException.h>
#include "../formal/FormalRegExpAlternation.h"
#include "../formal/FormalRegExpEmpty.h"
#include <sstream>

namespace regexp {

UnboundedRegExpAlternation::UnboundedRegExpAlternation ( ) {
}

UnboundedRegExpAlternation::UnboundedRegExpAlternation ( const UnboundedRegExpAlternation & other ) : elements ( other.elements ) {
}

UnboundedRegExpAlternation::UnboundedRegExpAlternation ( UnboundedRegExpAlternation && other ) noexcept : elements ( std::move ( other.elements ) ) {
	this->parentRegExp = other.parentRegExp;
	this->attachRegExp ( NULL );
}

UnboundedRegExpAlternation & UnboundedRegExpAlternation::operator =( const UnboundedRegExpAlternation & other ) {
	if ( this == & other ) return * this;

	* this = UnboundedRegExpAlternation ( other );

	return * this;
}

UnboundedRegExpAlternation & UnboundedRegExpAlternation::operator =( UnboundedRegExpAlternation && other ) noexcept {
	std::swap ( this->elements, other.elements );
	std::swap ( this->parentRegExp, other.parentRegExp ); // this->parentRegExp is stored within other.parentRegExp and it is reattached on the next line, done to save code (loop is in attachRegExp method)

	this->attachRegExp ( other.parentRegExp );

	return * this;
}

UnboundedRegExpAlternation::~UnboundedRegExpAlternation ( ) noexcept {
}

const std::vector < std::smart_ptr < const UnboundedRegExpElement > > & UnboundedRegExpAlternation::getElements ( ) const {
	return reinterpret_cast < const std::vector < std::smart_ptr < const UnboundedRegExpElement > > & > ( elements );
}

const std::vector < std::smart_ptr < UnboundedRegExpElement > > & UnboundedRegExpAlternation::getElements ( ) {
	return elements;
}

void UnboundedRegExpAlternation::appendElement ( UnboundedRegExpElement && element ) {
	this->elements.push_back ( std::smart_ptr < UnboundedRegExpElement > ( std::move ( element ).plunder ( ) ) );

	if ( this->parentRegExp && !elements.back ( )->attachRegExp ( this->parentRegExp ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void UnboundedRegExpAlternation::appendElement ( const UnboundedRegExpElement & element ) {
	appendElement ( std::move_copy ( element ) );
}

UnboundedRegExpElement * UnboundedRegExpAlternation::clone ( ) const {
	return new UnboundedRegExpAlternation ( * this );
}

UnboundedRegExpElement * UnboundedRegExpAlternation::plunder ( ) && {
	return new UnboundedRegExpAlternation ( std::move ( * this ) );
}

std::smart_ptr < FormalRegExpElement > UnboundedRegExpAlternation::asFormal ( ) const {
	if ( elements.size ( ) == 0 ) return std::smart_ptr < FormalRegExpElement > ( new FormalRegExpEmpty ( ) );

	std::smart_ptr < FormalRegExpElement > res = elements[elements.size ( ) - 1]->asFormal ( );

	for ( unsigned i = elements.size ( ) - 1; i >= 1; i-- )
		res = std::smart_ptr < FormalRegExpElement > ( new FormalRegExpAlternation ( std::move ( * elements[i - 1]->asFormal ( ) ), std::move ( * res ) ) );

	return res;
}

int UnboundedRegExpAlternation::compare ( const UnboundedRegExpAlternation & other ) const {
	std::compare < std::decay < decltype ( getElements ( ) ) >::type > comp;

	return comp ( getElements ( ), other.getElements ( ) );
}

void UnboundedRegExpAlternation::operator >>( std::ostream & out ) const {
	out << "(UnboundedRegExpAlternation";

	for ( const auto & e : elements )
		out << " " << * e;

	out << ")";
}

bool UnboundedRegExpAlternation::testSymbol ( const alphabet::Symbol & symbol ) const {
	for ( const auto & child : this->elements )
		if ( child->testSymbol ( symbol ) ) return true;

	return false;
}

bool UnboundedRegExpAlternation::attachRegExp ( const UnboundedRegExp * regexp ) {
	if ( this->parentRegExp == regexp ) return true;

	this->parentRegExp = regexp;

	for ( const auto & child : this->elements )
		if ( !child->attachRegExp ( regexp ) ) return false;

	return true;
}

void UnboundedRegExpAlternation::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	for ( const auto & child : this->elements )
		child->computeMinimalAlphabet ( alphabet );
}

UnboundedRegExpAlternation::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */
