/*
 * UnboundedRegExpConcatenation.cpp
 *
 *  Created on: Nov 27, 2013
 *      Author: Martin Zak
 */

#include "UnboundedRegExpConcatenation.h"
#include <exception/CommonException.h>
#include "../formal/FormalRegExpConcatenation.h"
#include "../formal/FormalRegExpEpsilon.h"
#include <sstream>

namespace regexp {

UnboundedRegExpConcatenation::UnboundedRegExpConcatenation ( ) {
}

UnboundedRegExpConcatenation::UnboundedRegExpConcatenation ( const UnboundedRegExpConcatenation & other ) : elements ( other.elements ) {
}

UnboundedRegExpConcatenation::UnboundedRegExpConcatenation ( UnboundedRegExpConcatenation && other ) noexcept : elements ( std::move ( other.elements ) ) {
	this->parentRegExp = other.parentRegExp;
	this->attachRegExp ( NULL );
}

UnboundedRegExpConcatenation & UnboundedRegExpConcatenation::operator =( const UnboundedRegExpConcatenation & other ) {
	if ( this == & other ) return * this;

	* this = UnboundedRegExpConcatenation ( other );

	return * this;
}

UnboundedRegExpConcatenation & UnboundedRegExpConcatenation::operator =( UnboundedRegExpConcatenation && other ) noexcept {
	std::swap ( this->elements, other.elements );
	std::swap ( this->parentRegExp, other.parentRegExp ); // this->parentRegExp is stored within other.parentRegExp and it is reattached on the next line, done to save code (loop is in attachRegExp method)

	this->attachRegExp ( other.parentRegExp );

	return * this;
}

UnboundedRegExpConcatenation::~UnboundedRegExpConcatenation ( ) noexcept {
}

const std::vector < std::smart_ptr < const UnboundedRegExpElement > > & UnboundedRegExpConcatenation::getElements ( ) const {
	return reinterpret_cast < const std::vector < std::smart_ptr < const UnboundedRegExpElement > > & > ( elements );
}

const std::vector < std::smart_ptr < UnboundedRegExpElement > > & UnboundedRegExpConcatenation::getElements ( ) {
	return elements;
}

void UnboundedRegExpConcatenation::appendElement ( UnboundedRegExpElement && element ) {
	this->elements.push_back ( std::smart_ptr < UnboundedRegExpElement > ( std::move ( element ).plunder ( ) ) );

	if ( this->parentRegExp && !elements.back ( )->attachRegExp ( this->parentRegExp ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void UnboundedRegExpConcatenation::appendElement ( const UnboundedRegExpElement & element ) {
	appendElement ( std::move_copy ( element ) );
}

UnboundedRegExpElement * UnboundedRegExpConcatenation::clone ( ) const {
	return new UnboundedRegExpConcatenation ( * this );
}

UnboundedRegExpElement * UnboundedRegExpConcatenation::plunder ( ) && {
	return new UnboundedRegExpConcatenation ( std::move ( * this ) );
}

std::smart_ptr < FormalRegExpElement > UnboundedRegExpConcatenation::asFormal ( ) const {
	if ( elements.size ( ) == 0 ) return std::smart_ptr < FormalRegExpElement > ( new FormalRegExpEpsilon ( ) );

	std::smart_ptr < FormalRegExpElement > res = elements[elements.size ( ) - 1]->asFormal ( );

	for ( unsigned i = elements.size ( ) - 1; i >= 1; i-- )
		res = std::smart_ptr < FormalRegExpElement > ( new FormalRegExpConcatenation ( std::move ( * elements[i - 1]->asFormal ( ) ), std::move ( * res ) ) );

	return res;
}

int UnboundedRegExpConcatenation::compare ( const UnboundedRegExpConcatenation & other ) const {
	std::compare < std::decay < decltype ( getElements ( ) ) >::type > comp;

	return comp ( getElements ( ), other.getElements ( ) );
}

void UnboundedRegExpConcatenation::operator >>( std::ostream & out ) const {
	out << "(UnboundedRegExpConcatenation";

	for ( const auto & e : elements )
		out << " " << * e;

	out << ")";
}

bool UnboundedRegExpConcatenation::testSymbol ( const alphabet::Symbol & symbol ) const {
	for ( const auto & child : this->elements )
		if ( child->testSymbol ( symbol ) ) return true;

	return false;
}

bool UnboundedRegExpConcatenation::attachRegExp ( const UnboundedRegExp * regexp ) {
	if ( this->parentRegExp == regexp ) return true;

	this->parentRegExp = regexp;

	for ( const auto & child : this->elements )
		if ( !child->attachRegExp ( regexp ) ) return false;

	return true;
}

void UnboundedRegExpConcatenation::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	for ( const auto & child : this->elements )
		child->computeMinimalAlphabet ( alphabet );
}

UnboundedRegExpConcatenation::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */
