/*
 * UnboundedRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_ELEMENT_H_
#define UNBOUNDED_REG_EXP_ELEMENT_H_

#include <XmlApi.hpp>
#include "../../alphabet/Symbol.h"
#include <set>

namespace regexp {

class UnboundedRegExp;

class UnboundedRegExpAlternation;
class UnboundedRegExpConcatenation;
class UnboundedRegExpIteration;
class UnboundedRegExpSymbol;
class UnboundedRegExpEmpty;
class UnboundedRegExpEpsilon;

class FormalRegExpElement;

class UnboundedRegExpElementVisitor {
public:
	virtual void Visit ( void *, const UnboundedRegExpAlternation & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpConcatenation & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpIteration & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpSymbol & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpEmpty & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpEpsilon & ) const = 0;
};

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
class UnboundedRegExpElement : public alib::CommonBase < UnboundedRegExpElement > {
protected:
	/**
	 * Parent regexp containing this instance of RegExpElement
	 */
	const UnboundedRegExp * parentRegExp;

	/**
	 * Attaches the regexp to this instance and all its childs
	 * @param regexp parent regexp to attach as parent
	 * @return true if symbols used in regexp element are in the regexp's alphabet
	 */
	virtual bool attachRegExp ( const UnboundedRegExp * regexp ) = 0;

public:
	virtual void Accept ( void * userData, const UnboundedRegExpElementVisitor & visitor ) const = 0;

	explicit UnboundedRegExpElement ( );

	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual std::smart_ptr < FormalRegExpElement > asFormal ( ) const = 0;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const = 0;

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	std::set < alphabet::Symbol > computeMinimalAlphabet ( ) const;

	friend class UnboundedRegExp;

	friend class UnboundedRegExpAlternation;
	friend class UnboundedRegExpConcatenation;
	friend class UnboundedRegExpIteration;
	friend class UnboundedRegExpSymbol;
	friend class UnboundedRegExpEmpty;
	friend class UnboundedRegExpEpsilon;
};

} /* namespace regexp */

namespace alib {

template < >
struct xmlApi < regexp::UnboundedRegExpElement > {
	static std::rvalue_ref < regexp::UnboundedRegExpElement > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const regexp::UnboundedRegExpElement & data );
};

} /* namespace alib */

#endif /* UNBOUNDED_REG_EXP_ELEMENT_H_ */
