/*
 * PrefixRankedTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixRankedTree.h"
#include "../TreeException.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "RankedTree.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../RankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

PrefixRankedTree::PrefixRankedTree ( std::set < alphabet::RankedSymbol > alphabet, std::vector < alphabet::RankedSymbol > data ) : std::Components < PrefixRankedTree, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ) {
	setContent ( std::move ( data ) );
}

PrefixRankedTree::PrefixRankedTree ( std::vector < alphabet::RankedSymbol > data ) : PrefixRankedTree ( std::set < alphabet::RankedSymbol > ( data.begin ( ), data.end ( ) ), std::move ( data ) ) {
}

PrefixRankedTree::PrefixRankedTree ( const RankedTree & tree ) : PrefixRankedTree ( tree.getAlphabet ( ), toPrefixRanked ( tree.getRoot ( ) ) ) {
}

std::vector < alphabet::RankedSymbol > PrefixRankedTree::toPrefixRanked ( const RankedNode & node ) {
	std::vector < alphabet::RankedSymbol > res;
	toPrefixRanked ( node, res );
	return res;
}

void PrefixRankedTree::toPrefixRanked ( const RankedNode & node, std::vector < alphabet::RankedSymbol > & linearNotation ) {
	linearNotation.push_back ( node.getSymbol ( ) );

	for ( const std::smart_ptr < const RankedNode > & child : node.getChildren ( ) )
		toPrefixRanked ( * child, linearNotation );
}

RankedTreeBase * PrefixRankedTree::clone ( ) const {
	return new PrefixRankedTree ( * this );
}

RankedTreeBase * PrefixRankedTree::plunder ( ) && {
	return new PrefixRankedTree ( std::move ( * this ) );
}

const std::vector < alphabet::RankedSymbol > & PrefixRankedTree::getContent ( ) const {
	return this->m_Data;
}

void PrefixRankedTree::setContent ( std::vector < alphabet::RankedSymbol > data ) {
	arityChecksum ( data );

	std::set < alphabet::RankedSymbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::RankedSymbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

void PrefixRankedTree::arityChecksum ( const std::vector < alphabet::RankedSymbol > & data ) {
	int arityChecksum = 1;

	for ( const alphabet::RankedSymbol & symbol : data ) {
		arityChecksum += symbol.getRank ( ).getData ( );
		arityChecksum -= 1;
	}

	if ( arityChecksum != 0 ) throw TreeException ( "The string does not form a tree" );
}

bool PrefixRankedTree::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int PrefixRankedTree::compare ( const PrefixRankedTree & other ) const {
	auto first = std::tie ( m_Data, getAlphabet() );
	auto second = std::tie ( other.m_Data, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void PrefixRankedTree::operator >>( std::ostream & out ) const {
	out << "(PrefixRankedTree ";

	for ( const alphabet::RankedSymbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

PrefixRankedTree::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::RankedSymbol & symbol : this->m_Data )
		ss << ( std::string ) symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

PrefixRankedTree PrefixRankedTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixRankedTree::getXmlTagName() );
	std::set < alphabet::RankedSymbol > rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet ( input );
	std::vector < alphabet::RankedSymbol > data = TreeFromXMLParser::parseRankedContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixRankedTree::getXmlTagName() );

	return PrefixRankedTree ( std::move ( rankedAlphabet ), std::move ( data ) );
}

void PrefixRankedTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixRankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet() );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixRankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace std {

template < >
bool tree::PrefixRankedTree::Component < tree::PrefixRankedTree, alphabet::RankedSymbol, tree::GeneralAlphabet >::used ( const alphabet::RankedSymbol & symbol ) const {
	const std::vector < alphabet::RankedSymbol > & content = static_cast < const tree::PrefixRankedTree * > ( this )->getContent ( );

	return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( );
}

template < >
bool tree::PrefixRankedTree::Component < tree::PrefixRankedTree, alphabet::RankedSymbol, tree::GeneralAlphabet >::available ( const alphabet::RankedSymbol & ) const {
	return true;
}

template < >
void tree::PrefixRankedTree::Component < tree::PrefixRankedTree, alphabet::RankedSymbol, tree::GeneralAlphabet >::valid ( const alphabet::RankedSymbol & ) const {
}

} /* namespace std */

namespace alib {

auto prefixRankedTreeParserRegister = xmlApi < tree::Tree >::ParserRegister < tree::PrefixRankedTree > ();
auto prefixRankedTreeParserRegister2 = xmlApi < tree::RankedTreeWrapper >::ParserRegister < tree::PrefixRankedTree > ();
auto prefixRankedTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < tree::PrefixRankedTree > ();

auto PrefixRankedTreeFromRankedTree = castApi::CastRegister < tree::PrefixRankedTree, tree::RankedTree > ( );
auto PrefixRankedTreeCastBinder = castApi::CastPoolStringBinder < tree::PrefixRankedTree > ( tree::PrefixRankedTree::getXmlTagName() );

} /* namespace alib */
