/*
 * RankedNode.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_NODE_H_
#define RANKED_NODE_H_

#include "../../alphabet/RankedSymbol.h"
#include <vector>
#include <set>
#include <XmlApi.hpp>

namespace tree {

class UnrankedNode;
class RankedTree;
class RankedPattern;
class RankedNonlinearPattern;

/**
 * Represents a node in the ranked tree. Contains name of the symbol.
 */
class RankedNode {
protected:
	alphabet::RankedSymbol symbol;

	std::vector < std::smart_ptr < RankedNode > > children;

	RankedNode * parent;

	/**
	 * Parent tree contanining this instance of RankedTree
	 */
	const std::set < alphabet::RankedSymbol > * alphabet;

	/**
	 * @copydoc RankedNode::attachTree()
	 */
	bool attachAlphabet ( const std::set < alphabet::RankedSymbol > * tree );

public:
	explicit RankedNode ( alphabet::RankedSymbol symbol, std::vector < std::smart_ptr < RankedNode > > children );

	RankedNode ( const RankedNode & other );
	RankedNode ( RankedNode && other ) noexcept;
	RankedNode & operator =( const RankedNode & other );
	RankedNode & operator =( RankedNode && other ) noexcept;
	~RankedNode ( ) noexcept;

	/**
	 * @copydoc RankedNode::cloneAsUnranked() const
	 */
	UnrankedNode asUnranked ( ) const;

	/**
	 * @copydoc RankedNode::testSymbol() const
	 */
	bool testSymbol ( const alphabet::RankedSymbol & symbol ) const;

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < alphabet::RankedSymbol > & alphabet ) const;

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	std::set < alphabet::RankedSymbol > computeMinimalAlphabet ( ) const;

	/**
	 * @return children
	 */
	const std::vector < std::smart_ptr < const RankedNode > > & getChildren ( ) const;

	/**
	 * @return children
	 */
	const std::vector < std::smart_ptr < RankedNode > > & getChildren ( );

	RankedNode * getParent ( );

	const RankedNode * getParent ( ) const;

	void setSymbol ( alphabet::RankedSymbol symbol );

	void swap ( RankedNode & other );

	int compare ( const RankedNode & ) const;

	/**
	 * @copydoc RankedNode::operator>>() const
	 */
	void operator >>( std::ostream & out ) const;

	friend std::ostream & operator <<( std::ostream &, const RankedNode & node );

	/**
	 * Returns the string representation of Node's Symbol.
	 */
	const alphabet::RankedSymbol & getSymbol ( ) const;

	explicit operator std::string ( ) const;

	void nicePrint ( std::ostream &, const std::string & = "", const bool = true ) const;

	friend class RankedTree;
	friend class RankedPattern;
	friend class RankedNonlinearPattern;
};

} /* namespace tree */

namespace alib {

template < >
struct xmlApi < tree::RankedNode > {
	static tree::RankedNode parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const tree::RankedNode & data );
};

} /* namespace alib */

namespace std {

template < >
struct compare < tree::RankedNode > {
	int operator ()( const tree::RankedNode & first, const tree::RankedNode & second ) const {
		return first.compare ( second );
	}

};

} /* namespace std */

#endif /* RANKED_NODE_H_ */
