/*
 * UnrankedPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNRANKED_PATTERN_H_
#define UNRANKED_PATTERN_H_

#include <string>
#include <set>
#include <core/components.hpp>

#include "UnrankedNode.h"
#include "../UnrankedTreeBase.h"

#include "../../alphabet/RankedSymbol.h"

namespace tree {

class UnrankedNode;
class RankedPattern;
class GeneralAlphabet;
class SubtreeWildcard;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
class UnrankedPattern : public UnrankedTreeBase, public std::Components < UnrankedPattern, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > {
	std::smart_ptr < UnrankedNode > pattern;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > alphabet, UnrankedNode pattern );
	explicit UnrankedPattern ( alphabet::Symbol subtreeWildcard, UnrankedNode pattern );
	explicit UnrankedPattern ( const RankedPattern & other );

	/**
	 * Copy constructor.
	 * @param other pattern to copy
	 */
	UnrankedPattern ( const UnrankedPattern & other );
	UnrankedPattern ( UnrankedPattern && other ) noexcept;
	UnrankedPattern & operator =( const UnrankedPattern & other );
	UnrankedPattern & operator =( UnrankedPattern && other ) noexcept;
	~UnrankedPattern ( ) noexcept;

	/**
	 * @return Root node of the regular expression pattern
	 */
	const UnrankedNode & getRoot ( ) const;

	virtual const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::Symbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const alphabet::Symbol & getSubtreeWildcard ( ) const {
		return accessElement < SubtreeWildcard > ( ).get ( );
	}

	/**
	 * @return Root node of the regular expression pattern
	 */
	UnrankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( UnrankedNode pattern );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnrankedPattern";

		return xmlTagName;
	}

	static UnrankedPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

#endif /* UNRANKED_PATTERN_H_ */
