/*
 * ReachableStates.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "ReachableStates.h"

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

namespace automaton {

namespace properties {

namespace efficient {

std::set<automaton::State> ReachableStates::reachableStates(const Automaton& automaton) {
	return dispatch(automaton.getData());
}

template<class T>
std::set<automaton::State> ReachableStates::reachableStates( const T & fsm ) {
	std::map<automaton::State, std::set<automaton::State>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	std::deque<automaton::State> queue { fsm.getInitialState( ) };
	std::set<automaton::State> visited { fsm.getInitialState( ) };

	while( !queue.empty() ) {
		const std::set<automaton::State>& to = transitions[queue.front()];
		queue.pop_front();

		for(const automaton::State& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<automaton::State&>(process)));
			}
	}

	return visited;
}

auto ReachableStatesEpsilonNFA = ReachableStates::RegistratorWrapper<std::set<automaton::State>, automaton::EpsilonNFA>(ReachableStates::reachableStates);
auto ReachableStatesNFA = ReachableStates::RegistratorWrapper<std::set<automaton::State>, automaton::NFA>(ReachableStates::reachableStates);
auto ReachableStatesCompactNFA = ReachableStates::RegistratorWrapper<std::set<automaton::State>, automaton::CompactNFA>(ReachableStates::reachableStates);
auto ReachableStatesExtendedNFA = ReachableStates::RegistratorWrapper<std::set<automaton::State>, automaton::ExtendedNFA>(ReachableStates::reachableStates);

template<>
std::set<automaton::State> ReachableStates::reachableStates( const automaton::MultiInitialStateNFA & fsm ) {
	std::map<automaton::State, std::set<automaton::State>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	std::deque<automaton::State> queue ( fsm.getInitialStates( ).begin(), fsm.getInitialStates().end() );
	std::set<automaton::State> visited = fsm.getInitialStates( );

	while( !queue.empty() ) {
		const std::set<automaton::State>& to = transitions[queue.front()];
		queue.pop_front();

		for(const automaton::State& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<automaton::State&>(process)));
			}
	}

	return visited;
}

auto ReachableStatesMultiInitialStateNFA = ReachableStates::RegistratorWrapper<std::set<automaton::State>, automaton::MultiInitialStateNFA>(ReachableStates::reachableStates);

template<>
std::set<automaton::State> ReachableStates::reachableStates( const automaton::DFA & fsm ) {
	std::map<automaton::State, std::set<automaton::State>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second);

	std::deque<automaton::State> queue { fsm.getInitialState( ) };
	std::set<automaton::State> visited { fsm.getInitialState( ) };

	while( !queue.empty() ) {
		const std::set<automaton::State>& to = transitions[queue.front()];
		queue.pop_front();

		for(const automaton::State& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<automaton::State&>(process)));
			}
	}

	return visited;
}

auto ReachableStatesDFA = ReachableStates::RegistratorWrapper<std::set<automaton::State>, automaton::DFA>(ReachableStates::reachableStates);

} /* namespace efficient */

} /* namespace properties */

} /* namespace automaton */

