/*
 * GraphFromStringParser.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#include "GraphFromStringParser.h"
#include "exception/CommonException.h"
#include "graph/GraphClasses.h"

#include "../StringApi.hpp"

namespace graph {

Graph GraphFromStringParser::parseGraph(std::istream &input) const
{
	GraphFromStringLexer::Token token = lexer.next(input);
	if (token.type != GraphFromStringLexer::TokenType::LPAR) {
		throw exception::CommonException("Invalid input. Expected LPAR");
	}

	const std::string &str = lexer.getString(input);
	if (str == "AdjacencyListDirectedGraph") {
		parseDelimiter(input); // :
		Graph graph{parseDirectedGraph<AdjacencyListDirectedGraph>(input)};
		if (lexer.next(input).type != GraphFromStringLexer::TokenType::RPAR) {
			throw exception::CommonException("Invalid input. Expected RPAR");
		}
		return graph;
	} else if (str == "AdjacencyMatrixDirectedGraph") {
		parseDelimiter(input); // :
		Graph graph{parseDirectedGraph<AdjacencyMatrixDirectedGraph>(input)};
		if (lexer.next(input).type != GraphFromStringLexer::TokenType::RPAR) {
			throw exception::CommonException("Invalid input. Expected RPAR");
		}
		return graph;
	} else if (str == "AdjacencyListUndirectedGraph") {
		parseDelimiter(input); // :
		Graph graph{parseUndirectedGraph<AdjacencyListUndirectedGraph>(input)};
		if (lexer.next(input).type != GraphFromStringLexer::TokenType::RPAR) {
			throw exception::CommonException("Invalid input. Expected RPAR");
		}
		return graph;
	} else if (str == "AdjacencyMatrixUndirectedGraph") {
		parseDelimiter(input); // :
		Graph graph{parseUndirectedGraph<AdjacencyMatrixUndirectedGraph>(input)};
		if (lexer.next(input).type != GraphFromStringLexer::TokenType::RPAR) {
			throw exception::CommonException("Invalid input. Expected RPAR");
		}
		return graph;
	}

	throw exception::CommonException("Invalid graph type");
}

template<typename GraphType>
GraphType GraphFromStringParser::parseDirectedGraph(std::istream &input) const
{
	GraphType graph;
	parseNodes(input, graph);
	parseDelimiter(input); // :
	parseDirectedEdges(input, graph);
	parseDelimiter(input); // :
	parseNodeValues(input, graph);
	parseDelimiter(input); // :
	parseDirectedEdgeValues(input, graph);
	return graph;
}
/*
AdjacencyMatrixDirectedGraph GraphFromStringParser::parseAMDirectedGraph(std::istream &input) const
{
	AdjacencyMatrixDirectedGraph graph;
	parseNodes(input, graph);
	parseDelimiter(input); // :
	parseDirectedEdges(input, graph);
	parseDelimiter(input); // :
	parseNodeValues(input, graph);
	parseDelimiter(input); // :
	parseDirectedEdgeValues(input, graph);
	return graph;
}
*/
template<typename GraphType>
GraphType GraphFromStringParser::parseUndirectedGraph(std::istream &input) const
{
	GraphType graph;
	parseNodes(input, graph);
	parseDelimiter(input); // :
	parseUndirectedEdges(input, graph);
	parseDelimiter(input); // :
	parseNodeValues(input, graph);
	parseDelimiter(input); // :
	parseUndirectedEdgeValues(input, graph);
	return graph;
}
/*
AdjacencyMatrixUndirectedGraph GraphFromStringParser::parseAMUndirectedGraph(std::istream &input) const
{
	AdjacencyMatrixUndirectedGraph graph;
	parseNodes(input, graph);
	parseDelimiter(input); // :
	parseUndirectedEdges(input, graph);
	parseDelimiter(input); // :
	parseNodeValues(input, graph);
	parseDelimiter(input); // :
	parseUndirectedEdgeValues(input, graph);
	return graph;
}
*/
REPRESENTATION GraphFromStringParser::parseRepresentation(std::istream &input) const
{
	std::string representation = lexer.getString(input);
	if (representation.empty()) {
		throw exception::CommonException("Invalid representation");
	}
	return representationFromString(representation);
}

Node GraphFromStringParser::parseNode(std::istream &input) const
{
	return Node(alib::stringApi<label::Label>::parse(input));
}

DirectedEdge GraphFromStringParser::parseDirectedEdge(std::istream &input) const
{
	Node from = parseNode(input);
	label::Label name = alib::stringApi<label::Label>::parse(input);
	Node to = parseNode(input);
	return DirectedEdge(from, to, name);
}

UndirectedEdge GraphFromStringParser::parseUndirectedEdge(std::istream &input) const
{
	Node first = parseNode(input);
	label::Label name = alib::stringApi<label::Label>::parse(input);
	Node second = parseNode(input);
	return UndirectedEdge(first, second, name);
}

template<typename T>
void GraphFromStringParser::parseNodes(std::istream &input, T &graph) const
{
	GraphFromStringLexer::Token token = lexer.next(input);

	// Empty nodes?
	if (token.type == GraphFromStringLexer::TokenType::COLON) {
		return;
	}
	lexer.putback(input, token);

	while (true) {
		graph.addNode(parseNode(input));

		token = lexer.next(input);
		if (token.type != GraphFromStringLexer::TokenType::COMMA) {
			lexer.putback(input, token);
			return;
		}
	}
}

void GraphFromStringParser::parseDirectedEdges(std::istream &input, DirectedGraph &graph) const
{
	GraphFromStringLexer::Token token = lexer.next(input);

	// Empty edges?
	if (token.type == GraphFromStringLexer::TokenType::COLON) {
		return;
	}
	lexer.putback(input, token);

	while (true) {
		graph.addEdge(parseDirectedEdge(input));

		token = lexer.next(input);
		if (token.type != GraphFromStringLexer::TokenType::COMMA) {
			lexer.putback(input, token);
			return;
		}
	}
}

void GraphFromStringParser::parseUndirectedEdges(std::istream &input, UndirectedGraph &graph) const
{
	GraphFromStringLexer::Token token = lexer.next(input);

	// Empty edges?
	if (token.type == GraphFromStringLexer::TokenType::COLON) {
		return;
	}
	lexer.putback(input, token);

	while (true) {
		graph.addEdge(parseUndirectedEdge(input));

		token = lexer.next(input);
		if (token.type != GraphFromStringLexer::TokenType::COMMA) {
			lexer.putback(input, token);
			return;
		}
	}
}

void GraphFromStringParser::parseDelimiter(std::istream &input) const
{
	GraphFromStringLexer::Token token = lexer.next(input);
	if (token.type != GraphFromStringLexer::TokenType::COLON) {
		throw exception::CommonException("Invalid input. Expected COLON");
	}
}

int GraphFromStringParser::parseValue(std::istream &input) const
{
	GraphFromStringLexer::Token token = lexer.next(input);
	if (token.type != GraphFromStringLexer::TokenType::EQUAL) {
		throw exception::CommonException("Invalid input. Expected EQUAL");
	}

	token = lexer.next(input);
	if (token.type != GraphFromStringLexer::TokenType::INTEGER) {
		throw exception::CommonException("Invalid input. Expected INTEGER");
	}

	return std::stoi(token.value);
}

template<typename T>
void GraphFromStringParser::parseNodeValues(std::istream &input, T &graph) const
{
	GraphFromStringLexer::Token token = lexer.peek(input);

	// Empty values?
	if (token.type == GraphFromStringLexer::TokenType::COLON) {
		return;
	}

	while (true) {
		graph::Node node = parseNode(input);
		int value = parseValue(input);
		graph.setNodeValue(node, value);

		token = lexer.next(input);
		if (token.type != GraphFromStringLexer::TokenType::COMMA) {
			lexer.putback(input, token);
			return;
		}
	}
}

void GraphFromStringParser::parseDirectedEdgeValues(std::istream &input, DirectedGraph &graph) const
{
	GraphFromStringLexer::Token token = lexer.peek(input);

	// Empty values?
	if (token.type == GraphFromStringLexer::TokenType::RPAR) {
		return;
	}

	while (true) {
		graph::DirectedEdge node = parseDirectedEdge(input);
		int value = parseValue(input);
		graph.setEdgeValue(node, value);

		token = lexer.next(input);
		if (token.type != GraphFromStringLexer::TokenType::COMMA) {
			lexer.putback(input, token);
			return;
		}
	}
}

void GraphFromStringParser::parseUndirectedEdgeValues(std::istream &input, UndirectedGraph &graph) const
{
	GraphFromStringLexer::Token token = lexer.peek(input);

	// Empty values?
	if (token.type == GraphFromStringLexer::TokenType::RPAR) {
		return;
	}

	while (true) {
		graph::UndirectedEdge node = parseUndirectedEdge(input);
		int value = parseValue(input);
		graph.setEdgeValue(node, value);

		token = lexer.next(input);
		if (token.type != GraphFromStringLexer::TokenType::COMMA) {
			lexer.putback(input, token);
			return;
		}
	}
}

} // namespace graph
