/*
 * astat.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>
#include <vector>
#include <sax/FromXMLParserHelper.h>

#include <factory/XmlDataFactory.hpp>
#include <exception/CommonException.h>

#include "AutomataStat.h"
#include "GrammarStat.h"
#include "RegExpStat.h"
#include "StringStat.h"
#include "TreeStat.h"
#include "SetStat.h"

PrintingOptions translatePrintingOptions ( std::string value ) {
	if ( value == "noop" )
		return PrintingOptions::NOOP;
	else if ( value == "print" )
		return PrintingOptions::PRINT;
	else if ( value == "size" )
		return PrintingOptions::SIZE;
	else if ( value == "both" )
		return PrintingOptions::BOTH;
	else
		throw exception::CommonException ( "Invalid Argument in translate printing options (" + value + ")" );
}

int main ( int argc, char * argv[] ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Prints usefull information about automata, grammars, regexps, ...", ' ', "0.01" );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > allowedPrintingOptions {
			"noop", "print", "size", "both"
		};
		TCLAP::ValuesConstraint < std::string > allowedPrintingOptionsVals ( allowedPrintingOptions );

		TCLAP::ValueArg < std::string > presentation ( "p", "presentation", "How to print data", true, "noop", & allowedPrintingOptionsVals );
		cmd.add ( presentation );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > automatonPrintingOptions {
			"states", "finalStates", "initialStates", "inputAlphabet", "transitions"
		};
		TCLAP::ValuesConstraint < std::string > automatonPrintingOptionsVals ( automatonPrintingOptions );

		TCLAP::MultiArg < std::string > automaton ( "", "automaton", "Display stats of automata", false, & automatonPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > grammarPrintingOptions {
			"nonterminals", "terminals", "initialSymbols", "rules"
		};
		TCLAP::ValuesConstraint < std::string > grammarPrintingOptionsVals ( grammarPrintingOptions );

		TCLAP::MultiArg < std::string > grammar ( "", "grammar", "Display stats of grammars", false, & grammarPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > regexpPrintingOptions {
			"alphabet", "nodes"
		};
		TCLAP::ValuesConstraint < std::string > regexpPrintingOptionsVals ( regexpPrintingOptions );

		TCLAP::MultiArg < std::string > regexp ( "", "regexp", "Display stats of regexps", false, & regexpPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > stringPrintingOptions {
			"alphabet", "content"
		};
		TCLAP::ValuesConstraint < std::string > stringPrintingOptionsVals ( stringPrintingOptions );

		TCLAP::MultiArg < std::string > string ( "", "string", "Display stats of strings", false, & stringPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > treePrintingOptions {
			"alphabet", "nodes"
		};
		TCLAP::ValuesConstraint < std::string > treePrintingOptionsVals ( treePrintingOptions );

		TCLAP::MultiArg < std::string > tree ( "", "tree", "Display stats of trees", false, & treePrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		TCLAP::SwitchArg set ( "", "set", "Display stats of sets", false );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < TCLAP::Arg * > xorlist;
		xorlist.push_back ( & automaton );
		xorlist.push_back ( & grammar );
		xorlist.push_back ( & regexp );
		xorlist.push_back ( & string );
		xorlist.push_back ( & tree );
		xorlist.push_back ( & set );
		cmd.xorAdd ( xorlist );

		TCLAP::ValueArg < std::string > file ( "i", "input", "Read from file", false, "-", "file" );
		cmd.add ( file );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if ( verbose.isSet ( ) )
			common::GlobalData::verbose = true;

		if ( measure.isSet ( ) )
			common::GlobalData::measure = true;

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Input read", measurements::Type::AUXILIARY );

		PrintingOptions printingOption = translatePrintingOptions ( presentation.getValue ( ) );

		std::deque < sax::Token > tokens = sax::FromXMLParserHelper::parseInput(file);

		if ( alib::XmlDataFactory::first < automaton::Automaton > ( tokens ) && automaton.isSet ( ) ) {
			AutomataSettings settings = { PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP };
			std::set < std::string > values {
				automaton.getValue ( ).begin ( ), automaton.getValue ( ).end ( )
			};

			if ( values.count ( "states" ) )
				settings.states = printingOption;

			if ( values.count ( "finalStates" ) )
				settings.finalStates = printingOption;

			if ( values.count ( "initialStates" ) )
				settings.initialStates = printingOption;

			if ( values.count ( "inputAlphabet" ) )
				settings.inputAlphabet = printingOption;

			if ( values.count ( "transitions" ) )
				settings.transitions = printingOption;

			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens < automaton::Automaton > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			AutomataStat::stat ( automaton, settings );
		} else if ( alib::XmlDataFactory::first < grammar::Grammar > ( tokens ) && grammar.isSet ( ) ) {
			GrammarSettings settings = { PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP };
			std::set < std::string > values {
				grammar.getValue ( ).begin ( ), grammar.getValue ( ).end ( )
			};

			if ( values.count ( "nonterminals" ) )
				settings.nonterminals = printingOption;

			if ( values.count ( "terminals" ) )
				settings.terminals = printingOption;

			if ( values.count ( "initialSymbols" ) )
				settings.initialSymbols = printingOption;

			if ( values.count ( "rules" ) )
				settings.rules = printingOption;

			grammar::Grammar grammar = alib::XmlDataFactory::fromTokens < grammar::Grammar > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			GrammarStat::stat ( grammar, settings );
		} else if ( alib::XmlDataFactory::first < regexp::RegExp > ( tokens ) && regexp.isSet ( ) ) {
			RegExpSettings settings = { PrintingOptions::NOOP, PrintingOptions::NOOP };
			std::set < std::string > values {
				regexp.getValue ( ).begin ( ), regexp.getValue ( ).end ( )
			};

			if ( values.count ( "alphabet" ) )
				settings.alphabet = printingOption;

			if ( values.count ( "nodes" ) )
				settings.nodes = printingOption;

			regexp::RegExp regexp = alib::XmlDataFactory::fromTokens < regexp::RegExp > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			RegExpStat::stat ( regexp, settings );
		} else if ( alib::XmlDataFactory::first < string::String > ( tokens ) && string.isSet ( ) ) {
			StringSettings settings = { PrintingOptions::NOOP, PrintingOptions::NOOP };
			std::set < std::string > values {
				string.getValue ( ).begin ( ), string.getValue ( ).end ( )
			};

			if ( values.count ( "alphabet" ) )
				settings.alphabet = printingOption;

			if ( values.count ( "content" ) )
				settings.content = printingOption;

			string::String string = alib::XmlDataFactory::fromTokens < string::String > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			StringStat::stat ( string, settings );
		} else if ( alib::XmlDataFactory::first < tree::Tree > ( tokens ) && tree.isSet ( ) ) {
			TreeSettings settings = { PrintingOptions::NOOP, PrintingOptions::NOOP };
			std::set < std::string > values {
				tree.getValue ( ).begin ( ), tree.getValue ( ).end ( )
			};

			if ( values.count ( "alphabet" ) )
				settings.alphabet = printingOption;

			if ( values.count ( "nodes" ) )
				settings.nodes = printingOption;

			tree::Tree tree = alib::XmlDataFactory::fromTokens < tree::Tree > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			TreeStat::stat ( tree, settings );
		} else if ( alib::XmlDataFactory::first < container::ObjectsSet > ( tokens ) && set.isSet ( ) ) {
			SetSettings settings = { PrintingOptions::NOOP };
			settings.general = printingOption;

			container::ObjectsSet data = alib::XmlDataFactory::fromTokens < container::ObjectsSet > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			SetStat::stat ( data, settings );
		} else if ( alib::XmlDataFactory::first < label::LabelSetLabel > ( tokens ) && set.isSet ( ) ) {
			SetSettings settings = { PrintingOptions::NOOP };
			settings.general = printingOption;

			label::LabelSetLabel data = alib::XmlDataFactory::fromTokens < label::LabelSetLabel > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			SetStat::stat ( data, settings );
		} else if ( alib::XmlDataFactory::first < alphabet::SymbolSetSymbol > ( tokens ) && set.isSet ( ) ) {
			SetSettings settings = { PrintingOptions::NOOP };
			settings.general = printingOption;

			alphabet::SymbolSetSymbol data = alib::XmlDataFactory::fromTokens < alphabet::SymbolSetSymbol > ( std::move ( tokens ) );

			measurements::end ( );
			measurements::start ( "Stats print", measurements::Type::MAIN );

			SetStat::stat ( data, settings );
		} else {
			throw exception::CommonException ( "Input not recognized." );
		}

		measurements::end ( );
		measurements::end ( );

		if ( measure.getValue ( ) ) std::cmeasure << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		std::cout << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		std::cerr << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
